from typing import overload

import numpy as np
import numpy.typing as npt

from .types import AnalyticFunc, ComplexScalarOrArray, ScalarOrArray


def central_diff(
    f: AnalyticFunc,
) -> AnalyticFunc:
    h = 1e-6

    @overload
    def df(
        z: npt.NDArray[np.complex_] | npt.NDArray[np.float_],
    ) -> ComplexScalarOrArray: ...

    @overload
    def df(z: complex | float) -> complex: ...

    def df(z: ScalarOrArray) -> ComplexScalarOrArray:
        return (f(z + h) - f(z - h)) / (2 * h)

    return df



import numpy as np

from .contour_interface import ContourABC
from .root_finding import RootFinderState, find_roots_gen
from .types import AnalyticFunc


def _contour_color(contour: ContourABC) -> tuple[float, float, float, float]:
    """
    Deterministically generate a colour for a contour so that the contour has the same
    colour in each frame of the root finding animation
    """
    import matplotlib.pyplot as plt

    cmap = plt.get_cmap("jet")

    rng = np.random.default_rng(hash(contour))
    return cmap(rng.random())


def _update_frame(frame: RootFinderState, original_contour: ContourABC) -> None:
    import matplotlib.pyplot as plt

    ax = plt.gca()

    roots, _, contour_data, num_remaining_roots = frame

    plt.cla()  # clear axis
    original_contour.plot(linecolor="k", linestyle="--")
    for data in contour_data:
        contour = data.contour
        plt.text(
            contour.central_point.real, contour.central_point.imag, str(data.num_roots)
        )
        contour.plot(linecolor=_contour_color(contour))

    plt.scatter(np.real(roots), np.imag(roots), color="k", marker="x")
    ax.text(
        0.02, 0.95, "Zeros remaining: %i" % num_remaining_roots, transform=ax.transAxes
    )
    original_contour.size_plot()
    plt.draw()


def demo_roots_animation(
    original_contour: ContourABC,
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    **roots_kwargs,
):
    """
    Creates an animation object where each frame is a step in the rootfinding process

    Parameters
    ----------
    original_contour : :class:`Contour <cxroots.contour.Contour>`
        The contour which bounds the region in which all the roots of
        f(z) are sought.
    f : function
        A function of a single complex variable, z, which is analytic
        within the contour and has no poles or roots on the contour.
    df : function, optional
        A function of a single complex variable which is the derivative
        of the function f(z). If df is not given then it will be
        approximated with a finite difference formula.
    **roots_kwargs : kwargs
        Other keyword arguments to pass to
        :func:`find_roots_gen <cxroots.root_finding.find_roots_gen>`

    Returns
    -------
    animation.FuncAnimation
        An animation where each frame is a step in the rootfinding process
    """
    import matplotlib.pyplot as plt
    from matplotlib import animation

    fig = plt.gcf()

    root_finder = find_roots_gen(original_contour, f, df, **roots_kwargs)
    return animation.FuncAnimation(
        fig,
        _update_frame,  # type: ignore
        frames=root_finder,  # type: ignore
        fargs=(original_contour,),
        cache_frame_data=False,
    )


def demo_find_roots(
    original_contour: ContourABC,
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    save_file: str | None = None,
    auto_animation: bool = False,
    writer: str | None = None,
    **roots_kwargs,
) -> None:
    """
    An animated demonstration of the root finding process using matplotlib.

    Parameters
    ----------
    save_file : str, optional
        If given then the animation will be saved to disk with filename
        equal to save_file instead of being shown.
    auto_animation : bool, optional
        If False (default) then press SPACE to step the animation forward
        If True then the animation will play automatically until all the
        roots have been found.
    writer : str, optional
        Passed to :meth:`matplotlib.animation.FuncAnimation.save`.
    **roots_kwargs
        Additional key word arguments passed to :meth:`~cxroots.contour.Contour.roots`.
    """
    import matplotlib.pyplot as plt

    if save_file or auto_animation:
        anim = demo_roots_animation(original_contour, f, df=None, **roots_kwargs)
        if save_file:
            anim.save(filename=save_file, fps=1, dpi=200, writer=writer)
            plt.close()
        else:
            plt.show()

    else:
        # Create event handler to let user move through frames
        root_finder = find_roots_gen(original_contour, f, df, **roots_kwargs)
        original_contour.plot(linecolor="k", linestyle="--")

        def draw_next(event):
            if event.key == " ":
                try:
                    _update_frame(next(root_finder), original_contour)
                except StopIteration:
                    # No more roots to find
                    pass

        fig = plt.gcf()
        fig.canvas.mpl_connect("key_press_event", draw_next)
        plt.show()



from abc import ABC, abstractmethod
from collections.abc import Generator, Sequence
from typing import Optional

from .types import AnalyticFunc, Color, IntegrationMethod


class ContourABC(ABC):
    @property
    @abstractmethod
    def central_point(self) -> complex:
        """A central point that lies within the contour"""
        ...

    @property
    @abstractmethod
    def area(self) -> float:
        """The area of the contour in the complex plane"""
        ...

    @abstractmethod
    def contains(self, z: complex) -> bool:
        """True if the point z is within the contour, false otherwise"""
        ...

    @abstractmethod
    def distance(self, z: complex) -> float:
        """
        The distance from the point z in the complex plane to the
        nearest point on the contour.
        """
        ...

    @abstractmethod
    def trap_product(
        self,
        k: int,
        f: AnalyticFunc,
        df: AnalyticFunc | None = None,
        phi: AnalyticFunc | None = None,
        psi: AnalyticFunc | None = None,
    ) -> complex:
        r"""
        Use Romberg integration to estimate the symmetric bilinear form used in
        (1.12) of [KB]_ using 2**k+1 samples

        .. math::

            <\phi,\psi> = \frac{1}{2\pi i} \oint_C \phi(z)\psi(z)\frac{f'(z)}{f(z)} dz
        """
        ...

    @abstractmethod
    def integrate(
        self,
        f: AnalyticFunc,
        abs_tol: float = 1.49e-08,
        rel_tol: float = 1.49e-08,
        div_max: int = 15,
        int_method: IntegrationMethod = "quad",
    ) -> complex:
        r"""
        Integrate the function f along the contour C

        .. math::

            \oint_C f(z) dz
        """
        ...

    @abstractmethod
    def count_roots(
        self,
        f: AnalyticFunc,
        df: AnalyticFunc | None = None,
        int_abs_tol: float = 0.07,
        integer_tol: float = 0.1,
        div_min: int = 3,
        div_max: int = 15,
        int_method: IntegrationMethod = "quad",
    ) -> int:
        """
        For a function of one complex variable, f(z), which is analytic in
        and within the contour C, return the number of zeros (counting
        multiplicities) within the contour
        """
        ...

    @abstractmethod
    def subdivisions(
        self, axis: str = "alternating"
    ) -> Generator[tuple["ContourABC", ...], None, None]:
        """A generator for possible subdivisions of the contour"""
        ...

    @property
    @abstractmethod
    def parent(self) -> Optional["ContourABC"]:
        """The contour that this contour was created from as part of subdivision"""
        ...

    @property
    @abstractmethod
    def children(self) -> Sequence["ContourABC"] | None:
        """The contours that were created from this contour during subdivision"""
        ...

    @abstractmethod
    def plot(
        self, num_points: int = 100, linecolor: Color = "C0", linestyle: str = "-"
    ) -> None:
        """
        Uses matplotlib to plot, but not show, the contour as a 2D plot in
        the Complex plane.
        """
        ...

    @abstractmethod
    def size_plot(self) -> None:
        """Adjust the plot axes limits to nicely frame the contour"""
        ...



from .contours.annulus import Annulus
from .contours.annulus_sector import AnnulusSector
from .contours.circle import Circle
from .contours.rectangle import Rectangle
from .version import __version__  # noqa:F401

# Define public interface
__all__ = [
    "Annulus",
    "AnnulusSector",
    "Circle",
    "Rectangle",
]



# have a common seed for each testing process
from math import pi

import numpy as np
import pytest

from cxroots.paths import ComplexArc, ComplexLine

rng = np.random.default_rng(159)


@pytest.mark.parametrize(
    "a", rng.uniform(-10, 10, size=3) + 1j * rng.uniform(-10, 10, size=3)
)
@pytest.mark.parametrize(
    "b", rng.uniform(-10, 10, size=3) + 1j * rng.uniform(-10, 10, size=3)
)
@pytest.mark.parametrize(
    "p", rng.uniform(-10, 10, size=3) + 1j * rng.uniform(-10, 10, size=3)
)
def test_distance_line(a, b, p):
    t = np.linspace(0, 1, 100001)
    line = ComplexLine(a, b)
    assert line.distance(p) == pytest.approx(np.min(np.abs(line(t) - p)), abs=1e-6)


@pytest.mark.parametrize(
    "z0", rng.uniform(-10, 10, size=3) + 1j * rng.uniform(-10, 10, size=3)
)
@pytest.mark.parametrize("r", rng.uniform(0, 10, size=3))
@pytest.mark.parametrize("t0", rng.uniform(0, 2 * pi, size=3))
@pytest.mark.parametrize("dt", rng.uniform(-2 * pi, 2 * pi, size=3))
@pytest.mark.parametrize(
    "p", rng.uniform(-10, 10, size=3) + 1j * rng.uniform(-10, 10, size=3)
)
def test_distance_arc(z0, r, t0, dt, p):
    z0 = 0
    t = np.linspace(0, 1, 100001)
    arc = ComplexArc(z0, r, t0, dt)
    assert arc.distance(p) == pytest.approx(np.min(np.abs(arc(t) - p)), abs=1e-6)



from typing import NamedTuple

import numpy as np

from cxroots.contour_interface import ContourABC


class RootResult(
    NamedTuple("RootResult", [("roots", list[complex]), ("multiplicities", list[int])])
):
    """
    A class which stores the roots and their multiplicites as attributes
    and provides convienent methods for displaying them.

    Attributes
    ----------
    roots : list
        List of roots
    multiplicities : list
        List of multiplicities where the ith element of the list is the
        multiplicity of the ith element of roots.
    contour : Contour
        The contour bounding the region in which the roots were found.
    """

    def __new__(
        cls, roots: list[complex], multiplicities: list[int], contour: ContourABC
    ):
        return super().__new__(cls, roots, multiplicities)

    def __init__(
        self, roots: list[complex], multiplicities: list[int], contour: ContourABC
    ):
        self.contour = contour
        super().__init__()

    def show(self, save_file: str | None = None) -> None:
        """
        Plot the roots and the initial integration contour in the
        complex plane.

        Parameters
        ----------
        save_file : str, optional
            If provided the plot of the roots will be saved with
            file name save_file instead of being shown.

        Example
        -------
        .. plot::
            :include-source:

            from cxroots import Circle
            C = Circle(0, 2)
            f = lambda z: z**6 + z**3
            df = lambda z: 6*z**5 + 3*z**2
            r = C.roots(f, df)
            r.show()
        """
        import matplotlib.pyplot as plt

        self.contour.plot(linecolor="k", linestyle="--")
        plt.scatter(np.real(self.roots), np.imag(self.roots), color="k", marker="x")

        if save_file is not None:
            plt.savefig(save_file)
            plt.close()
        else:
            plt.show()

    def __str__(self):
        roots, multiplicities = np.array(self.roots), np.array(self.multiplicities)

        # reorder roots
        sortargs = np.argsort(roots)
        roots, multiplicities = roots[sortargs], multiplicities[sortargs]

        s = " Multiplicity |               Root              "
        s += "\n------------------------------------------------"

        for i, root in np.ndenumerate(roots):
            if root.real < 0:
                s += (
                    f"\n{int(multiplicities[i]): ^14d}| "
                    f"{root.real:.12f} {root.imag:+.12f}i"
                )
            else:
                s += (
                    f"\n{int(multiplicities[i]): ^14d}|  "
                    f"{root.real:.12f} {root.imag:+.12f}i"
                )

        return s



from numpydoc.docscrape import FunctionDoc


def remove_para(*paras):
    def wrapper(func):
        doc = FunctionDoc(func)
        for i, p in enumerate(doc["Parameters"][:]):
            if p.name.split(":")[0].rstrip() in paras:
                del doc["Parameters"][i]
        func.__doc__ = doc
        return func

    return wrapper


def update_docstring(**dic):
    def wrapper(func):
        doc = FunctionDoc(func)
        for k, v in dic.items():
            doc[k] = v
        func.__doc__ = doc
        return func

    return wrapper


class NumberOfRootsChangedError(Exception):
    pass



"""
References
----------
[DSZ] "Locating all the Zeros of an Analytic Function in one Complex Variable"
    M.Dellnitz, O.Schutze, Q.Zheng, J. Compu. and App. Math. (2002), Vol.138, Issue 2
[DL] "A Numerical Method for Locating the Zeros of an Analytic function",
    L.M.Delves, J.N.Lyness, Mathematics of Computation (1967), Vol.21, Issue 100
[KB] "Computing the zeros of analytic functions" by Peter Kravanja, Marc Van Barel,
    Springer 2000
"""

import cmath
import unittest

import numpy as np
import pytest
from numpy import cos, exp, sin, sqrt

from cxroots import Annulus, Circle, Rectangle
from cxroots.tests.approx_equal import roots_approx_equal


class RootfindingTests:
    # These attributes need to be defined by test case subclasses
    C = None
    f = None
    df = None
    roots = None
    multiplicities = None

    def test_rootfinding_romb_df(self):
        roots_approx_equal(
            self.C.roots(self.f, self.df, int_method="romb"),
            (self.roots, self.multiplicities),
            decimal=10,
        )

    def test_rootfinding_romb_f(self):
        roots_approx_equal(
            self.C.roots(self.f, int_method="romb"),
            (self.roots, self.multiplicities),
            decimal=10,
        )

    def test_rootfinding_quad_df(self):
        roots_approx_equal(
            self.C.roots(self.f, self.df, int_method="quad"),
            (self.roots, self.multiplicities),
            decimal=10,
        )

    def test_rootfinding_quad_f(self):
        roots_approx_equal(
            self.C.roots(self.f, int_method="quad"),
            (self.roots, self.multiplicities),
            decimal=10,
        )


class TestRootfindingNoRoots(unittest.TestCase, RootfindingTests):
    def setUp(self):
        self.C = Annulus(1, [1, 2])
        self.f = lambda z: (z - 1) ** 3
        self.df = lambda z: 3 * (z - 1) ** 2

        self.roots = []
        self.multiplicities = []


class TestRootfindingPoly1(unittest.TestCase, RootfindingTests):
    def setUp(self):
        self.C = Rectangle([-2, 2], [-2, 2])
        self.f = lambda z: z**3 * (z - 1.2) ** 2
        self.df = lambda z: 3 * (z) ** 2 * (z - 1.2) ** 2 + 2 * z**3 * (z - 1.2)

        self.roots = [0, 1.2]
        self.multiplicities = [3, 2]


class TestRootfindingPoly2(unittest.TestCase, RootfindingTests):
    def setUp(self):
        self.C = Annulus(0, [0.5, 2.5])
        self.f = lambda z: (z - 2) ** 2 * (z - 1) ** 5
        self.df = lambda z: 2 * (z - 2) * (z - 1) ** 5 + 5 * (z - 2) ** 2 * (z - 1) ** 4

        self.roots = [1, 2]
        self.multiplicities = [5, 2]


class TestRootfinding141(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.1 from [KB]
        self.C = Circle(0, 3)
        e = 1e-2
        self.f = lambda z: (z - e) * (1 + (z - sqrt(3)) ** 2)
        self.df = lambda z: (1 + (z - sqrt(3)) ** 2) + (z - e) * 2 * (z - sqrt(3))

        self.roots = [e, sqrt(3) + 1j, sqrt(3) - 1j]
        self.multiplicities = [1, 1, 1]


class TestRootfinding142(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.2 from [KB]
        self.C = Circle(0, 2)
        self.f = lambda z: exp(3 * z) + 2 * z * cos(z) - 1
        self.df = lambda z: 3 * exp(3 * z) + 2 * cos(z) - 2 * z * sin(z)

        self.roots = [
            0,
            -1.844233953262213,
            0.5308949302929305 + 1.33179187675112098j,
            0.5308949302929305 - 1.33179187675112098j,
        ]
        self.multiplicities = [1, 1, 1, 1]


class TestRootfinding142b(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.2 from [KB] with a rectangular initial contour
        self.C = Rectangle([-2, 2], [-2, 2])
        self.f = lambda z: exp(3 * z) + 2 * z * cos(z) - 1
        self.df = lambda z: 3 * exp(3 * z) + 2 * cos(z) - 2 * z * sin(z)

        self.roots = [
            0,
            -1.844233953262213,
            0.5308949302929305 + 1.33179187675112098j,
            0.5308949302929305 - 1.33179187675112098j,
        ]
        self.multiplicities = [1, 1, 1, 1]


class TestRootfinding143(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.3 from [KB]
        self.C = Circle(0, 5)
        self.f = lambda z: z**2 * (z - 1) * (z - 2) * (z - 3) * (z - 4) + z * sin(z)
        self.df = (
            lambda z: 2 * z * (3 * z**4 - 25 * z**3 + 70 * z**2 - 75 * z + 24)
            + sin(z)
            + z * cos(z)
        )

        self.roots = [
            0,
            1.18906588973011365517521756,
            1.72843498616506284043592924,
            3.01990732809571222812005354,
            4.03038191606046844562845941,
        ]
        self.multiplicities = [2, 1, 1, 1, 1]


class TestRootfinding144(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.4 from [KB]
        self.C = Circle(0, 3)
        self.f = lambda z: (z * (z - 2)) ** 2 * (
            exp(2 * z) * cos(z) + z**3 - 1 - sin(z)
        )
        self.df = (
            lambda z: 2 * z * (z - 2) ** 2 * (exp(2 * z) * cos(z) + z**3 - 1 - sin(z))
            + 2 * (z - 2) * z**2 * (exp(2 * z) * cos(z) + z**3 - 1 - sin(z))
            + (z * (z - 2)) ** 2
            * (2 * exp(2 * z) * cos(z) - exp(2 * z) * sin(z) + 3 * z**2 - cos(z))
        )

        self.roots = [
            -0.4607141197289707542294459477 - 0.6254277693477682516688207854j,
            -0.4607141197289707542294459477 + 0.6254277693477682516688207854j,
            0,
            2,
            1.66468286974551654134568653,
        ]
        self.multiplicities = [1, 1, 3, 2, 1]


class TestRootfinding145(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.5 from [KB]
        self.C = Circle(0, 11)
        self.f = lambda z: np.prod([z - k for k in range(1, 11)], axis=0)
        self.df = lambda z: np.sum(
            [
                np.prod([z - k for k in range(1, 11) if k != m], axis=0)
                for m in range(1, 11)
            ],
            axis=0,
        )

        self.roots = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.multiplicities = np.ones(10)


class TestRootfinding145b(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.4.5 from [KB] with a rectangular initial contour
        self.C = Rectangle([-1, 11], [-1, 1])
        self.f = lambda z: np.prod([z - k for k in range(1, 11)], axis=0)
        self.df = lambda z: np.sum(
            [
                np.prod([z - k for k in range(1, 11) if k != m], axis=0)
                for m in range(1, 11)
            ],
            axis=0,
        )

        self.roots = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.multiplicities = np.ones(10)


class TestRootfinding151(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.5.1 from [KB]
        self.C = Rectangle([-2, 2], [-2, 3])
        self.f = lambda z: exp(3 * z) + 2 * z * cos(z) - 1
        self.df = lambda z: 3 * exp(3 * z) + 2 * cos(z) - 2 * z * sin(z)

        self.roots = np.array(
            [
                -1.84423395326221337491592440,
                0,
                0.5308949302929305274642203840 - 1.331791876751120981651544228j,
                0.5308949302929305274642203840 + 1.331791876751120981651544228j,
            ],
            dtype=complex,
        )
        self.multiplicities = [1, 1, 1, 1]

    def test_rootfinding_b_df(self):
        roots_approx_equal(
            self.C.roots(self.f, self.df, M=2),
            (self.roots, self.multiplicities),
            decimal=12,
        )

    def test_rootfinding_b_f(self):
        roots_approx_equal(
            self.C.roots(self.f, M=2),
            (self.roots, self.multiplicities),
            decimal=12,
        )


class TestRootfinding152(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.5.2 from [KB]
        self.C = Rectangle([-0.5, 5.5], [-0.5, 1.5])
        self.f = lambda z: z**2 * (z - 1) * (z - 2) * (z - 3) * (z - 4) + z * sin(z)
        self.df = (
            lambda z: 2 * z * (3 * z**4 - 25 * z**3 + 70 * z**2 - 75 * z + 24)
            + sin(z)
            + z * cos(z)
        )

        self.roots = [
            0,
            1.18906588973011365517521756,
            1.72843498616506284043592924,
            3.01990732809571222812005354,
            4.03038191606046844562845941,
        ]
        self.multiplicities = [2, 1, 1, 1, 1]


class TestRootfinding153(unittest.TestCase, RootfindingTests):
    def setUp(self):
        # Ex 1.5.3 from [KB]
        self.C = Rectangle([-1, 3], [-1, 1])
        self.f = lambda z: (z * (z - 2)) ** 2 * (
            exp(2 * z) * cos(z) + z**3 - 1 - sin(z)
        )
        self.df = (
            lambda z: 2 * z * (z - 2) ** 2 * (exp(2 * z) * cos(z) + z**3 - 1 - sin(z))
            + 2 * z**2 * (z - 2) * (exp(2 * z) * cos(z) + z**3 - 1 - sin(z))
            + (z * (z - 2)) ** 2
            * (2 * exp(2 * z) * cos(z) - exp(2 * z) * sin(z) + 3 * z**2 - cos(z))
        )

        self.roots = [
            0,
            2,
            1.66468286974551654134568653,
            -0.4607141197289707542294459477 - 0.6254277693477682516688207854j,
            -0.4607141197289707542294459477 + 0.6254277693477682516688207854j,
        ]
        self.multiplicities = [3, 2, 1, 1, 1]


def test_reevaluation_of_num_roots():
    contour = Circle(0, 2)

    def f(z):
        return (z - 1) * (z - 0.2) ** 2

    roots = [1, 0.2]
    multiplicities = [1, 2]
    roots_approx_equal(
        contour.roots(f, int_abs_tol=10, int_method="romb"),
        (roots, multiplicities),
    )


class TestIntroduction(unittest.TestCase, RootfindingTests):
    def setUp(self):
        self.C = Circle(0, 3)
        self.f = lambda z: (z * (z + 2)) ** 2 * (
            exp(2 * z) * cos(z) - 1 - sin(z) + z**5
        )
        self.df = lambda z: 2 * (exp(2 * z) * cos(z) - 1 - sin(z) + z**5) * (
            z**2 * (z + 2) + (z + 2) ** 2 * z
        ) + (z * (z + 2)) ** 2 * (
            2 * exp(2 * z) * cos(z) - exp(2 * z) * sin(z) - cos(z) + 5 * z**4
        )

        self.roots = [
            0,
            -2,
            2.23755778246706002284084684,
            -0.6511140702635986824274097994 - 0.3904257190882864369857773146j,
            -0.6511140702635986824274097994 + 0.3904257190882864369857773146j,
            0.64857808095387581293067569277 - 1.35662268398824203963215495605j,
            0.64857808095387581293067569277 + 1.35662268398824203963215495605j,
        ]
        self.multiplicities = [3, 2, 1, 1, 1, 1, 1]


def test_annular_combustion():
    A = -0.19435  # noqa: N806
    B = 1000.41  # noqa: N806
    C = 522463  # noqa: N806
    T = 0.005  # noqa: N806

    def f(z):
        return z**2 + A * z + B * exp(-T * z) + C

    def df(z):
        return 2 * z + A - B * T * exp(-T * z)

    rectangle = Rectangle([-15000, 5000], [-15000, 15000])

    import warnings

    warnings.filterwarnings("error")
    roots = rectangle.roots(f, df, root_err_tol=1e-6)
    assert len(roots.roots) == 24


@pytest.mark.parametrize("int_method", ["quad", "romb"])
def test_const_df(int_method):
    def f(z):
        return z - 0.5

    def df(z):
        return 1

    contour = Circle(0, 1)
    roots = contour.roots(f, df, int_method=int_method)

    assert roots.roots == [0.5]
    assert roots.multiplicities == [1]


@pytest.mark.parametrize("int_method", ["quad", "romb"])
def test_df(int_method):
    def f(z):
        return (z - 0.5) ** 2

    def df(z):
        return 2 * (z - 0.5)

    contour = Circle(0, 1)
    roots = contour.roots(f, df, int_method=int_method)

    assert roots.roots == pytest.approx([0.5])
    assert roots.multiplicities == [2]


def test_rootfinding_299():
    """
    Regression test for https://github.com/rparini/cxroots/issues/299
    In this case there were some roots missing and some with more multiplicites than
    there should have been.
    """
    # Geometry Details
    l = 1.0  # noqa: E741
    a = 0.25
    b = l - a

    # Region 1 (unburnt gas)
    T1 = 300  # noqa: N806
    P1 = 101325  # noqa: N806
    R = 287.0  # noqa: N806 Gas constant
    rho1 = P1 / (R * T1)  # Density of the unburnt gas (kg/m3)
    gamma = 1.4
    C1 = np.sqrt(gamma * R * T1)  # noqa: N806 Speed of sound

    # Region 2 (burnt gas)
    T2 = 1200  # noqa: N806
    P2 = P1  # noqa: N806
    rho2 = P2 / (R * T2)
    C2 = np.sqrt(gamma * R * T2)  # noqa: N806

    zeta = rho1 * C1 / (rho2 * C2)

    # Crocco's n-tau model
    n = 10.0
    tau = 0.005

    def f(omega):
        return zeta * cmath.cos(omega / C1 * a) * cmath.cos(omega / C2 * b) - (
            1 + n * cmath.exp(1j * omega * tau)
        ) * (cmath.sin(omega / C1 * a) * cmath.sin(omega / C2 * b))

    def f_prime(omega):  # Derivative of f
        return (
            -zeta
            * (
                b / C2 * cmath.cos(omega * a / C1) * cmath.sin(omega * b / C2)
                + a / C1 * cmath.cos(omega * b / C2) * cmath.sin(omega * a / C1)
            )
            - (1 + n * cmath.exp(1j * omega * tau))
            * (
                b / C2 * cmath.sin(omega * a / C1) * cmath.cos(omega * b / C2)
                + a / C1 * cmath.sin(omega * b / C2) * cmath.cos(omega * a / C1)
            )
            - (1j * n * tau * cmath.exp(1j * omega * tau))
            * (cmath.sin(omega * a / C1) * cmath.sin(omega * b / C2))
        )

    # Using a rectangle as the contour
    contour = Rectangle([0, 5000], [-500, 500])
    roots, multiplicities = contour.roots(f, f_prime)
    roots = sorted(roots, key=lambda r: r.real)

    expected_roots = [
        261.045540365941 - 181.036716404397j,
        904.889245400156 + 343.599454434591j,
        1880.005303030232 + 388.327867463358j,
        2801.720372617268 + 150.680161943082j,
        3112.682726748578 + 47.264321292923j,
        4362.901996620711 + 0.000000000000j,
        4398.645160420987 + 184.305444695056j,
    ]
    assert all(m == 1 for m in multiplicities)
    assert roots == pytest.approx(expected_roots)


if __name__ == "__main__":
    unittest.main(verbosity=3)



import functools
import logging
from typing import overload

import numpy as np
import numpy.typing as npt
import scipy.linalg

from .contour_interface import ContourABC
from .root_counting import RombCallback, prod
from .types import AnalyticFunc, ComplexScalarOrArray, IntegrationMethod, ScalarOrArray


def approximate_roots(
    C: ContourABC,  # noqa: N803
    N: int,  # noqa: N803
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    abs_tol: float = 1.49e-08,
    rel_tol: float = 1.49e-08,
    err_stop: float = 1e-10,
    div_min: int = 3,
    div_max: int = 15,
    root_tol: float = 1e-8,
    int_method: IntegrationMethod = "quad",
    callback: RombCallback | None = None,
) -> tuple[tuple[complex, ...], tuple[float, ...]]:
    """
    Approximate the roots and multiplcities of the function f within the
    contour C using the method of [KB]_.  The multiplicites are computed
    using eq. (21) in [SLV]_.

    Parameters
    ----------
    C : :class:`~<cxroots.contour.Contour>`
        The contour which encloses the roots of f the user wishes to find.
    N : int
        The number of roots (counting multiplicties) of f within C.
        This is the result of calling :meth:`~cxroots.contour.Contour.count_roots`.
    f : function
        The function for which the roots are sought.  Must be a function
        of a single complex variable, z, which is analytic within C and
        has no poles or roots on the C.
    df : function, optional
        A function of a single complex variable which is the derivative
        of the function f(z). If df is not given then it will be
        approximated with a finite difference formula.
    abs_tol : float, optional
        Absolute error tolerance for integration.
    rel_tol : float, optional
        Relative error tolerance for integration.
    err_stop : float, optional
        The number of distinct roots within a contour, n, is determined
        by checking if all the elements of a list of contour integrals
        involving formal orthogonal polynomials are sufficently close to
        zero, ie. that the absolute value of each element is < err_stop.
        If err_stop is too large/small then n may be smaller/larger than
        it actually is.
    div_min : int, optional
        If the Romberg integration method is used then div_min is the
        minimum number of divisions before the Romberg integration
        routine is allowed to exit.
    div_max : int, optional
        If the Romberg integration method is used then div_max is the
        maximum number of divisions before the Romberg integration
        routine exits.
    root_tol : float, optional
        If any roots are within root_tol of one another then they will be
        treated as duplicates and removed.  This helps to alleviate the
        problem of err_stop being too small.
    int_method : {'quad', 'romb'}, optional
        If 'quad' then :func:`scipy.integrate.quad` is used to perform
        integration.  If 'romb' then Romberg integraion is performed
        instead.
    callback : function, optional
        Only used if int_method is 'romb'. Passed to
        :func:`~<cxroots.root_counting.prod>`.

    Returns
    -------
    tuple of complex
        The distinct roots of f within the contour C.
    tuple of float
        The corresponding multiplicites of the roots within C.  Should
        be integers but will not be automatically rounded here.

    References
    ----------
    .. [KB] P. Kravanja and M. Van Barel. "Computing the Zeros of
        Anayltic Functions". Springer (2000)
    .. [SLV] E. Strakova, D. Lukas, P. Vodstrcil. "Finding Zeros of
        Analytic Functions and Local Eigenvalue Analysis Using Contour
        Integral Method in Examples". Mathematical Analysis and Numerical
        Mathematics, Vol. 15, 2, (2017)
    """
    logger = logging.getLogger(__name__)
    logger.info("Approximating the " + str(N) + " roots in: " + str(C))

    if N == 0:
        return (), ()

    product = functools.partial(
        prod,
        C,
        f,
        df,
        abs_tol=abs_tol,
        rel_tol=rel_tol,
        div_min=div_min,
        div_max=div_max,
        int_method=int_method,
        callback=callback,
    )

    s = [N, product(lambda z: z)]  # ordinary moments
    mu = s[1] / N
    phi_zeros = [np.array([]), np.array([mu])]

    def phi(i: int) -> AnalyticFunc:
        if len(phi_zeros[i]) == 0:
            return lambda z: complex(1)
        coeff = np.poly(phi_zeros[i])
        # We should be using Polynomial.fromroots but this is not hashable so
        # causes problems with caching
        return lambda z: np.polyval(coeff, z)
        # return npp.Polynomial.fromroots(phi_zeros[i])

    def phi1phi(i: int) -> AnalyticFunc:
        @overload
        def func(z: complex | float) -> complex: ...

        @overload
        def func(
            z: npt.NDArray[np.complex_] | npt.NDArray[np.float_],
        ) -> npt.NDArray[np.complex_] | complex: ...

        def func(z: ScalarOrArray) -> ComplexScalarOrArray:
            return phi(1)(z) * phi(i)(z)

        return func

    # initialize G_{pq} = <phi_p, phi_q>
    G = np.zeros((N, N), dtype=np.complex128)  # noqa: N806
    G[0, 0] = N  # = <phi_0, phi_0> = <1,1>

    # initialize G1_{pq} = <phi_p, phi_1 phi_q>
    G1 = np.zeros((N, N), dtype=np.complex128)  # noqa: N806
    G1[0, 0] = 0  # = <phi_0, phi_1 phi_0> = <1, z-mu> = s1-mu*N = 0

    r, t = 1, 0
    while r + t < N:
        ### define FOP of degree r+t+1
        p = r + t
        G[p, 0 : p + 1] = [product(phi(p), phi(q)) for q in range(r + t + 1)]
        G[0 : p + 1, p] = G[p, 0 : p + 1]  # G is symmetric
        logger.debug("G=\n" + str(G[: p + 1, : p + 1]))

        G1[p, 0 : p + 1] = [product(phi(p), phi1phi(q)) for q in range(r + t + 1)]
        G1[0 : p + 1, p] = G1[p, 0 : p + 1]  # G1 is symmetric
        logger.debug("G1=\n" + str(G1[: p + 1, : p + 1]))

        """
        If any of the zeros of the FOP are outside of the interior
        of the contour then we assume that they are 'arbitary' and
        instead define the FOP as an inner polynomial. [KB]
        """
        poly_roots = scipy.linalg.eig(G1[: p + 1, : p + 1], G[: p + 1, : p + 1])[0] + mu
        if np.all([C.contains(z) for z in poly_roots]):
            r, t = r + t + 1, 0
            phi_zeros.append(poly_roots)
            logger.debug(
                "Regular polynomial " + str(r + t) + " roots: " + str(phi_zeros[-1])
            )

            # is the number of distinct roots, n=r?
            phi_func_last = phi(-1)
            for j in range(N - r):
                ip = product(lambda z: phi_func_last(z) * (z - mu) ** j, phi_func_last)
                logger.debug("%i of %i, abs(ip)=%f" % (j, N - r, abs(ip)))
                if abs(ip) > err_stop:
                    # n != r so carry on
                    logger.debug("n != " + str(r))
                    break
            else:
                # the for loop did not break
                logger.debug("n = " + str(r))
                break

        else:
            # define an inner polynomial as phi_{r+t+1} = phi_{t+1} phi_{r}
            t += 1
            phi_zeros.append(np.append(phi_zeros[t], phi_zeros[r]))
            logger.debug(
                "Inner polynomial " + str(r + t) + " roots: " + str(phi_zeros[-1])
            )

    roots = phi_zeros[-1]

    # remove any roots which are not distinct
    roots_to_remove = []
    for i, root in enumerate(roots):
        if len(roots[i + 1 :]) > 0 and np.any(np.abs(root - roots[i + 1 :]) < root_tol):
            roots_to_remove.append(i)
    roots = np.delete(roots, roots_to_remove)
    n = len(roots)

    ### compute the multiplicities, eq. (1.19) in [KB]
    # V = np.column_stack([roots**i for i in range(n)])
    # if n > 2: logger.debug('Computing ordinary moments')
    # s += [product(lambda z: z**p)[0] for p in range(2, n)]
    # multiplicities = np.dot(s[:n], np.linalg.inv(V))

    ### compute the multiplicities, eq. (21) in [SLV]
    V = np.array([[phi(j)(root) for root in roots] for j in range(n)])  # noqa: N806
    multiplicities = np.dot(np.linalg.inv(V), G[:n, 0])

    ### The method used in the vandermonde module doesn't seem significantly
    ### better than np.dot(s, np.linalg.inv(V)).  Especially since we know
    ### the result must be an integer anyway.
    # import vandermonde
    # multiplicities_vandermonde = vandermonde.solve_transpose(
    #     np.array(roots), np.array(s)
    # )

    ### Note that n = rank(H_N) is not used since calculating the
    ### rank of a matrix of floats can be quite unstable
    # s_func = lambda p: prod(C, f, df, lambda z: z**p)[0]
    # HN = np.fromfunction(np.vectorize(lambda p,q: s_func(p+q)), shape=(N,N))
    # print('n?', np.linalg.matrix_rank(HN, tol=1e-10))

    logger.debug(
        f"Approximate (roots, multiplicities): {list(zip(roots, multiplicities))}"
    )
    return tuple(roots), tuple(multiplicities)



import numpy as np


def sets_approx_equal(a, b, decimal=10):
    """
    Test if iterables a and b are approximately the same, up to reordering.

    Works by greedily reordering a and b so that they are the closest to
    each other element-wise and then using np.testing.assert_almost_equal

    .. np.testing.assert_almost_equal:
       https://docs.scipy.org/doc/numpy/reference/generated/numpy.testing.assert_almost_equal.html
    """

    a, b = np.array(a), np.array(b)

    # greedily reorder b so that it is closest to a element-wise
    for ai, a_element in np.ndenumerate(a):
        # find the element of b closest to a_element
        bi = np.argmin(np.abs(b - a_element))
        # put the element of b closest to a_element in the same place as a_element
        b[ai], b[bi] = b[bi], b[ai]

    # use numpy to compare the two arrays element-wise
    return np.testing.assert_almost_equal(a, b, decimal)


def roots_approx_equal(a, b, decimal=10):
    a_roots, a_multiplicities = a
    b_roots, b_multiplicities = b

    a_roots, b_roots = np.array(a_roots), np.array(b_roots)
    a_multiplicities, b_multiplicities = (
        np.array(a_multiplicities),
        np.array(b_multiplicities),
    )

    for ai, a_element in np.ndenumerate(a_roots):
        bi = np.argmin(np.abs(b_roots - a_element))
        b_roots[ai], b_roots[bi] = b_roots[bi], b_roots[ai]
        b_multiplicities[ai], b_multiplicities[bi] = (
            b_multiplicities[bi],
            b_multiplicities[ai],
        )

    np.testing.assert_almost_equal(a_roots, b_roots, decimal)
    np.testing.assert_almost_equal(a_multiplicities, b_multiplicities, decimal)



from math import pi
from typing import Literal

import numpy as np

from ..contour import Contour
from ..paths import ComplexArc, ComplexLine


class AnnulusSector(Contour):
    """
    A sector of an annulus in the complex plane.

    Parameters
    ----------
    center : complex
        The center of the annulus sector.
    radii : tuple
        Tuple of length two of the form (inner_radius, outer_radius)
    phi_range : tuple
        Tuple of length two of the form (phi0, phi1).
        The segment of the contour containing inner and outer circular
        arcs will be joined, counter clockwise from phi0 to phi1.

    Examples
    --------
    .. plot::
        :include-source:

        from numpy import pi
        from cxroots import AnnulusSector
        annulusSector = AnnulusSector(
            center=0.2, radii=(0.5, 1.25), phi_range=(-pi/4, pi/4)
        )
        annulusSector.show()

    .. plot::
        :include-source:

        from numpy import pi
        from cxroots import AnnulusSector
        annulusSector = AnnulusSector(
            center=0.2, radii=(0.5, 1.25), phi_range=(pi/4, -pi/4)
        )
        annulusSector.show()
    """

    axis_names = ("r", "phi")

    def __init__(
        self,
        center: complex,
        radii: tuple[float, float],
        phi_range: tuple[float, float],
    ):
        self.center = center

        if phi_range[0] > phi_range[1]:
            phi_range = (phi_range[0], phi_range[1] + 2 * pi)

        phi0, phi1 = self.phi_range = phi_range

        # r > 0
        r0, r1 = self.radii = radii
        if r0 < 0 or r1 <= 0:
            raise ValueError("Radius > 0")

        # verticies [[radius0,phi0],[radius0,phi1],[radius1,phi1],[radius0,phi1]]
        self.z1 = z1 = center + r0 * np.exp(1j * phi0)
        self.z2 = z2 = center + r1 * np.exp(1j * phi0)
        self.z3 = z3 = center + r1 * np.exp(1j * phi1)
        self.z4 = z4 = center + r0 * np.exp(1j * phi1)

        segments = [
            ComplexLine(z1, z2),
            ComplexArc(center, r1, phi0, phi1 - phi0),
            ComplexLine(z3, z4),
            ComplexArc(center, r0, phi1, phi0 - phi1),
        ]

        super().__init__(segments)

    def __str__(self):
        return (
            f"Annulus sector: center={self.center.real:.3f}{self.center.imag:+.3f}i, "
            f"r0={self.radii[0]:.3f}, r1={self.radii[1]:.3f}, "
            f"phi0={self.phi_range[0]:.3f}, phi1={self.phi_range[1]:.3f}"
        )

    @property
    def central_point(self) -> complex:
        # get the central point within the contour
        r = (self.radii[0] + self.radii[1]) / 2
        phi = (self.phi_range[0] + self.phi_range[1]) / 2
        return r * np.exp(1j * phi)

    @property
    def area(self) -> float:
        return (
            (self.radii[1] ** 2 - self.radii[0] ** 2)
            * abs(self.phi_range[1] - self.phi_range[0])
            % (2 * pi)
            / 2
        )

    def contains(self, z: complex) -> bool:
        """Returns True if the point z lies within the contour, False if otherwise"""
        angle = float(np.angle(z - self.center)) % (2 * pi)  # np.angle maps to [-pi,pi]
        radius_correct = self.radii[0] < abs(z - self.center) < self.radii[1]

        phi = np.mod(self.phi_range, 2 * pi)
        if phi[0] > phi[1]:
            angle_correct = (phi[0] < angle <= 2 * pi) or (0 <= angle < phi[1])
        else:
            angle_correct = phi[0] < angle < phi[1]

        return radius_correct and angle_correct

    def subdivide(
        self, axis: Literal["r", "phi"], division_factor: float = 0.5
    ) -> tuple["AnnulusSector", "AnnulusSector"]:
        """
        Subdivide the contour

        Parameters
        ----------
        axis : str, can be either 'r' or 'phi'
            The axis along which the line subdividing the contour is a constant.
        division_factor : float in range (0,1), optional
            Determines the point along 'axis' at which the line dividing the box is
            placed

        Returns
        -------
        box1 : AnnulusSector
            If axis is 'r' then phi_range and the inner radius is the same as original
            AnnulusSector with the outer radius determined by the division_factor.
            If axis is 'phi' then the radii and phi_range[0] is the same as the original
            AnnulusSector with phi_range[1] determined by the division_factor.
        box2 : AnnulusSector
            If axis is 'r' then phi_range and the outer radius is the same as original
            AnnulusSector with the inner radius determined equal to the outer radius
            of box1. If axis is 'phi' then the radii and phi_range[1] is the same as
            the original AnnulusSector with phi_range[0] equal to phi_range[1] of box1.
        """
        r0, r1 = self.radii
        phi0, phi1 = self.phi_range

        if axis == "r":
            division_point = r0 + division_factor * (r1 - r0)
            box1 = AnnulusSector(self.center, (r0, division_point), self.phi_range)
            box2 = AnnulusSector(self.center, (division_point, r1), self.phi_range)

            # reuse line segments from original box where possible
            # this allows the cached integrals to be used
            box1.segments[3] = self.segments[3]
            box2.segments[1] = self.segments[1]
            box1.segments[1]._reverse_path = box2.segments[3]
            box2.segments[3]._reverse_path = box1.segments[1]

        elif axis == "phi":
            division_point = phi0 + division_factor * (phi1 - phi0)
            box1 = AnnulusSector(self.center, self.radii, (phi0, division_point))
            box2 = AnnulusSector(self.center, self.radii, (division_point, phi1))

            box1.segments[0] = self.segments[0]
            box2.segments[2] = self.segments[2]
            box1.segments[2]._reverse_path = box2.segments[0]
            box2.segments[0]._reverse_path = box1.segments[2]

        else:
            raise ValueError("axis must be 'r' or 'phi'")

        for box in [box1, box2]:
            box._created_by_subdivision_axis = axis
            box._parent = self
        self._children = [box1, box2]

        return box1, box2



import functools
import logging
import warnings
from collections.abc import Callable, Generator, Iterable
from dataclasses import dataclass
from typing import Any, NamedTuple

import numpy as np
from numpydoc.docscrape import FunctionDoc
from rich.progress import BarColumn, Progress, TextColumn

from .contour_interface import ContourABC
from .iterative_methods import iterate_to_root
from .root_approximation import approximate_roots
from .root_counting import RootError
from .root_result import RootResult
from .types import AnalyticFunc, IntegrationMethod
from .util import NumberOfRootsChangedError, update_docstring


class MultiplicityError(RuntimeError):
    pass


class CountCalls:
    """
    Count how many times a given function is called.
    """

    def __init__(self, func: Callable[[Any], Any]):
        self.func = func
        self.calls = 0
        self.points = 0

    def __call__(self, z: Any) -> Any:
        self.calls += 1
        if hasattr(z, "__len__"):
            self.points += len(z)
        else:
            self.points += 1
        return self.func(z)


@dataclass
class ContourData:
    contour: ContourABC
    num_roots: int


def make_contour_data(
    contours: list[ContourABC], num_roots: dict[ContourABC, int]
) -> list[ContourData]:
    return [ContourData(contour=c, num_roots=num_roots[c]) for c in contours]


class RootFinderState(NamedTuple):
    roots: list[complex]
    multiplicities: list[int]
    contour_data: list[ContourData]
    num_remaining_roots: int


def find_roots_gen(
    original_contour: ContourABC,
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    guess_roots: Iterable[complex] = [],
    guess_roots_symmetry: Callable[[complex], Iterable[complex]] | None = None,
    newton_step_tol: float = 1e-14,
    refine_roots_beyond_tol: bool = True,
    newton_max_iter: int = 50,
    root_err_tol: float = 1e-10,
    abs_tol: float = 1.49e-08,
    rel_tol: float = 1.49e-08,
    integer_tol: float = 0.1,
    int_abs_tol: float = 0.07,
    M: int = 5,  # noqa: N803
    err_stop: float = 1e-10,
    int_method: IntegrationMethod = "quad",
    div_min: int = 3,
    div_max: int = 15,
) -> Generator[RootFinderState, None, None]:
    """
    A generator which at each step takes a contour and either finds all
    the zeros of f within it or subdivides it further.  Based on the
    algorithm in [KB]_.

    Parameters
    ----------
    original_contour : :class:`Contour <cxroots.contour.Contour>`
        The contour which bounds the region in which all the roots of
        f(z) are sought.
    f : function
        A function of a single complex variable, z, which is analytic
        within the contour and has no poles or roots on the contour.
    df : function, optional
        A function of a single complex variable which is the derivative
        of the function f(z). If df is not given then it will be
        approximated with a finite difference formula.
    guess_roots : list, optional
        A list of known roots or guesses for roots (they are checked
        before being accepted).
    guess_roots_symmetry : function, optional
        A function of a single complex variable, z, which returns a list
        of all points which are expected to be roots of f, given that z
        is a root of f.
    newton_step_tol : float, optional
        The required accuracy of the root.  The iterative method used to
        give a final value for each root will exit if the step size, dx,
        between sucessive iterations satisfies abs(dx) < newton_step_tol
        and refine_roots_beyond_tol is False.
    refine_roots_beyond_tol : bool, optional
        If True then the iterative method used to refine the roots will
        exit when error of the previous iteration, x0, was at least as
        good as the current iteration, x, in the sense that
        abs(f(x)) >= abs(f(x0)) and the previous iteration satisfied
        abs(dx0) < newton_step_tol.  In this case the preivous iteration
        is returned as the approximation of the root.
    newton_max_iter : int, optional
        The iterative method used to give a final value for each root
        will exit if the number of iterations exceeds newton_max_iter.
    root_err_tol : float, optional
        A complex value z is considered a root if abs(f(z)) < root_err_tol
    abs_tol : float, optional
        Absolute error tolerance used by the contour integration.
    rel_tol : float, optional
        Relative error tolerance used by the contour integration.
    integer_tol : float, optional
        A number is considered an integer if it is within integer_tol of
        an integer.  Used when determing if the value for the number of
        roots within a contour and the values of the computed
        multiplicities of roots are acceptably close to integers.
    int_abs_tol : float, optional
        The absolute error tolerance used for the contour integration
        when determining the number of roots within a contour.  Since
        the result of this integration must be an integer it can be much
        less accurate than usual.
    M : int, optional
        If the number of roots (including multiplicites) within a contour
        is greater than M then the contour is subdivided further.  M must
        be greater than or equal to the largest multiplcity of any root.
    err_stop : float, optional
        The number of distinct roots within a contour, n, is determined
        by checking if all the elements of a list of contour integrals
        involving formal orthogonal polynomials are sufficently close to
        zero, ie. that the absolute value of each element is < err_stop.
        If err_stop is too large/small then n may be smaller/larger than
        it actually is.
    int_method : {'quad', 'romb'}, optional
        If 'quad' then :func:`scipy.integrate.quad` is used to perform the
        integral.  If 'romb' then Romberg integraion, using
        :func:`scipy.integrate.romb`, is performed instead.  Typically, quad is
        the better choice but it requires that the real and imaginary
        parts of each integral are calculated sepeartely, in addition,
        if df is not provided, 'quad' will require additional function
        evaluations to approximate df at each point that f is evaluated
        at.  If evaluating f is expensive then 'romb' may be more
        efficient since it computes the real and imaginary parts
        simultaniously and if df is not provided it will approximate it
        using only the values of f that would be required by the
        integration routine in any case.
    div_min : int, optional
        If the Romberg integration method is used then div_min is the
        minimum number of divisions before the Romberg integration
        routine is allowed to exit.
    div_max : int, optional
        If the Romberg integration method is used then div_max is the
        maximum number of divisions before the Romberg integration
        routine exits.

    Yields
    ------
    list[complex]
        Roots of f(z) within the contour original_contour
    list[int]
        Multiplicites of roots
    list[(contour, roots_in_contour)]
        The contours which still contain roots
    int
        Remaining number of roots to be found within the contour

    References
    ----------
    .. [KB] Peter Kravanja, Marc Van Barel, "Computing the Zeros of
        Anayltic Functions", Springer (2000)
    """
    from .contours.circle import Circle

    # wrap f to record the number of function calls
    f = CountCalls(f)

    count_kwargs = {
        "f": f,
        "df": df,
        "int_abs_tol": int_abs_tol,
        "integer_tol": integer_tol,
        "div_min": div_min,
        "div_max": div_max,
        "int_method": int_method,
    }

    logger = logging.getLogger(__name__)
    roots: list[complex] = []
    multiplicities: list[int] = []
    failed_contours: list[ContourABC] = []
    contours: list[ContourABC] = []
    num_roots: dict[ContourABC, int] = {}

    try:
        # compute the total number of zeros, including multiplicities, within the
        # originally given contour
        num_roots[original_contour] = original_contour.count_roots(**count_kwargs)
        contours.append(original_contour)
    except RuntimeError as e:
        raise RuntimeError(
            """
            Integration along the initial contour has failed. There is likely a root on
            or close to the initial contour. Try changing the initial contour, if
            possible. Changing the integration method (int_method) may also help.
            """
        ) from e

    def subdivide(parent: ContourABC) -> None:
        """Given a contour, parent_contour, subdivide it into multiple contours."""
        logger.info("Subdividing " + str(parent))

        for subcontours in parent.subdivisions():
            # if a contour has already been used and caused an error then skip it
            failed_contour_desc = list(map(str, failed_contours))
            if any(
                contour_desc in failed_contour_desc
                for contour_desc in list(map(str, subcontours))
            ):
                continue

            # if a contour is near to a known root then skip this subdivision
            for root in roots:
                if any(contour.distance(root) < 0.01 for contour in subcontours):
                    continue

            try:
                subcontour_num_roots = [
                    contour.count_roots(**count_kwargs)
                    for contour in np.array(subcontours)
                ]
                while (
                    num_roots[parent] != sum(subcontour_num_roots)
                    and int_abs_tol > 0.01
                ):
                    logger.warning(
                        "Number of roots in sub contours not adding up to parent "
                        "contour. Recomputing number of roots in parent and child "
                        f"contours with int_abs_tol={0.5 * int_abs_tol}"
                    )
                    temp_count_kwargs = count_kwargs.copy()
                    temp_count_kwargs["int_abs_tol"] *= 0.5
                    num_roots[parent] = parent.count_roots(**temp_count_kwargs)
                    subcontour_num_roots = [
                        contour.count_roots(**temp_count_kwargs)
                        for contour in np.array(subcontours)
                    ]

                if num_roots[parent] == sum(subcontour_num_roots):
                    break

            except RootError:
                # If the number of zeros within either of the new contours is not an
                # integer then it is likely that the introduced line which subdivides
                # 'parent_contour' lies on a zero. To avoid this we will try to place
                # the subdividing line at a different point along the division axis
                logger.warning(
                    "RootError encountered when subdivding "
                    + str(parent)
                    + " into:\n"
                    + str(subcontours[0])
                    + "\n"
                    + str(subcontours[1])
                )
                continue
        else:
            # The list of possible subdivisions has been exhaused
            raise RuntimeError(
                f"""Unable to subdivide contour:
                \t{parent}
                """
            )

        # add sub-contours with roots to the list of contours to find the roots in
        for n_roots, c in zip(subcontour_num_roots, subcontours):
            if n_roots != 0:
                contours.append(c)
                num_roots[c] = n_roots

    def reverse_subdivision(contour: ContourABC) -> None:
        """
        Remove the contour and all its siblings and children from the list of contours
        to be examined/subdivided and put its parent back in the list to be subdivided
        """
        if contour.parent is None:
            raise ValueError("Contour has no parent so subdivision cannot be reversed")

        relations = []
        if contour.parent.children is not None:
            # The contour has siblings
            relations.extend([c for c in contour.parent.children if c != contour])

        if contour.children is not None:
            relations.extend(contour.children)

        for c in contours.copy():
            if c in relations:
                contours.remove(c)

        # put the parent contour back into the list of contours to subdivide again
        contours.append(contour.parent)

        # do not use this contour again
        failed_contours.append(contour)

    def add_root(root: complex, multiplicity: int) -> None:
        # check that the root we have found is distinct from the ones we already have
        if not roots or np.all(abs(np.array(roots) - root) > newton_step_tol):
            # add the root to the list if it is within the original contour
            if original_contour.contains(root):
                roots.append(root)
                multiplicities.append(multiplicity)
                logger.info(
                    "Recorded root "
                    + str(root)
                    + " with multiplicity "
                    + str(multiplicity)
                )
            else:
                logger.debug(
                    "Root " + str(root) + " ignored as not within original contour"
                )

            # check to see if there are any other roots implied by the given symmetry
            if guess_roots_symmetry is not None:
                for x0 in guess_roots_symmetry(root):
                    # first check that x0 is distinct from the roots we already have
                    if np.all(abs(np.array(roots) - x0) > newton_step_tol):
                        logger.info(str(root) + " is a root so checking " + str(x0))
                        new_root = iterate_to_root(
                            x0,
                            f,
                            df,
                            newton_step_tol,
                            root_err_tol,
                            newton_max_iter,
                            refine_roots_beyond_tol,
                        )
                        if new_root is not None:
                            circle = Circle(new_root, 1e-3)
                            contours.append(circle)
                            num_roots[circle] = circle.count_roots(**count_kwargs)
        else:
            logger.debug("Already recorded root " + str(root))

    # Add contours surrounding known roots so that they will be checked
    for root in guess_roots:
        circle = Circle(root, 1e-3)
        num_roots[circle] = circle.count_roots(**count_kwargs)
        contours.append(circle)

    while contours:
        # yield the initial state here so that the animation in demo_find_roots shows
        # the first frame
        num_found_roots = sum(multiplicities)
        remaining_roots = num_roots[original_contour] - num_found_roots
        yield RootFinderState(
            roots,
            multiplicities,
            make_contour_data(contours, num_roots),
            remaining_roots,
        )
        contour = contours.pop()

        # if a known root is too near to this contour then reverse the subdivision that
        # created it
        if np.any([contour.distance(root) < newton_step_tol for root in roots]):
            reverse_subdivision(contour)
            continue

        # if the contour is smaller than the newton_step_tol then just assume that
        # the root is at the center of the contour, print a warning and move on
        if contour.area < newton_step_tol:
            root = iterate_to_root(
                contour.central_point,
                f,
                df,
                newton_step_tol,
                root_err_tol,
                newton_max_iter,
                refine_roots_beyond_tol,
            )
            if (
                root is None
                or abs(f(root)) > abs(f(contour.central_point))
                or not contour.contains(root)
            ):
                root = contour.central_point

            warnings.warn(
                "The area of this contour is smaller than newton_step_tol. Try "
                f"increasing root_tol. The point z = {root.real} + {root.imag}i has "
                f"been recorded as a root of multiplicity {num_roots[contour]}. "
                f"The error |f(z)| = {abs(f(root))}"
            )
            add_root(root, num_roots[contour])
            continue

        # if all the roots within the contour have been located then continue to the
        # next contour
        num_known_roots = sum(
            multiplicity
            for root, multiplicity in zip(roots, multiplicities)
            if contour.contains(root)
        )
        if num_roots[contour] == num_known_roots:
            continue

        # if there are too many roots within the contour then subdivide or
        # if there are any known roots within the contour then also subdivide
        # (so as not to waste time re-approximating these roots)
        if num_known_roots > 0 or num_roots[contour] > M:
            subdivide(contour)
            continue

        # Approximate the roots in this contour
        if int_method == "romb":
            # Check to see if the number of roots has changed after new values of f
            # have been sampled
            new_count_kwargs = count_kwargs.copy()

            def callback(integral, err, num_div):
                if num_div > new_count_kwargs["div_min"]:
                    logger.info(
                        "Checking root count using the newly sampled values of f"
                    )
                    new_count_kwargs["div_min"] = num_div
                    new_num_roots = contour.count_roots(**new_count_kwargs)

                    if new_num_roots != num_roots[contour]:
                        logger.info(
                            "Number of roots has been recalculated using more samples"
                            " of f"
                        )
                        original_num_roots = num_roots[contour]
                        num_roots[contour] = new_num_roots
                        raise NumberOfRootsChangedError(
                            "The additional function evaluations of f taken while "
                            "approximating the roots within the contour have "
                            "shown that the number of roots of f within the contour "
                            f"is {new_num_roots} instead of {original_num_roots}."
                        )

        else:

            def callback(integral, err, num_div):
                # Don't do anything on callback
                pass

        try:
            approx_roots, approx_multiplicities = approximate_roots(
                contour,
                num_roots[contour],
                f,
                df,
                abs_tol=abs_tol,
                rel_tol=rel_tol,
                err_stop=err_stop,
                div_min=div_min,
                div_max=div_max,
                root_tol=newton_step_tol,
                int_method=int_method,
                callback=callback,
            )
        except NumberOfRootsChangedError:
            logger.debug("The number of roots within the contour has been reevaluated")
            if num_roots[contour] > M:
                subdivide(contour)
            else:
                contours.append(contour)
            continue

        if any(
            abs(round(m.real) - m.real) > integer_tol
            or abs(m.imag) > integer_tol
            or m < 1
            for m in approx_multiplicities
        ):
            logger.debug(
                f"Subdividing due to invalid multiplicies: {approx_multiplicities}"
            )
            subdivide(contour)
            continue

        for approx_root, approx_multiplicity in zip(
            approx_roots, approx_multiplicities
        ):
            # attempt to refine the root
            multiplicity = round(approx_multiplicity.real)
            root = iterate_to_root(
                approx_root,
                f,
                df,
                newton_step_tol,
                root_err_tol,
                newton_max_iter,
                refine_roots_beyond_tol,
            )

            if root is None or abs(f(approx_root)) < abs(f(root)):
                # stick with the original approximation
                root = approx_root

            if abs(f(root)) < root_err_tol:
                if np.any(
                    np.abs(np.round(approx_multiplicities) - approx_multiplicities)
                    > integer_tol
                ):
                    # the computed multiplicity might be unreliable so make a contour
                    # focused on that point instead so that the multiplicity will be
                    # recomputed. If the current contour is already centered on that
                    # point then make one with a smaller radius
                    radius = (
                        min(contour.radius * 0.1, 1e-3)
                        if isinstance(contour, Circle) and contour.center == root
                        else 1e-3
                    )
                    circle = Circle(root, radius)
                    contours.append(circle)
                    num_roots[circle] = circle.count_roots(**count_kwargs)
                else:
                    add_root(root, multiplicity)

            # if the root turns out to be very close to the contour then this may have
            # introduced an error.  Therefore, compute the multiplicity of this root
            # directly and disregard this contour (repeat its parent's subdivision).
            if contour.distance(root) < newton_step_tol:
                # remove the contour and any relations
                reverse_subdivision(contour)

        # if we haven't found all the roots then subdivide further
        num_known_roots = sum(
            multiplicity
            for root, multiplicity in zip(roots, multiplicities)
            if contour.contains(root)
        )
        if num_roots[contour] != num_known_roots and contour not in failed_contours:
            subdivide(contour)

    result = RootResult(roots, multiplicities, original_contour)
    logger.info(
        "Completed rootfinding with "
        + str(f.calls)
        + " evaluations of the given analytic function at "
        + str(f.points)
        + " points\n"
        + str(result)
    )

    num_found_roots = sum(multiplicities)
    remaining_roots = num_roots[original_contour] - num_found_roots
    yield RootFinderState(
        roots, multiplicities, make_contour_data(contours, num_roots), remaining_roots
    )


@update_docstring(Parameters=FunctionDoc(find_roots_gen)["Parameters"])
@functools.wraps(find_roots_gen, assigned=("__module__", "__name__"))
def find_roots(
    original_contour: ContourABC,
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    verbose: bool = False,
    **kwargs,
) -> RootResult:
    """
    Find all the roots of the complex analytic function f within the
    given contour.

    Parameters
    ----------
    %(find_roots_gen.parameters)s
    verbose : bool, optional
        If True print a progress bar showing the rootfinding progress.

    Returns
    -------
    result : :class:`RootResult <cxroots.root_result.RootResult>`
        A container for the roots and their multiplicities.
    """
    roots, multiplicities = [], []
    root_finder = find_roots_gen(original_contour, f, df, **kwargs)

    if verbose:
        text_column = TextColumn("{task.description}")
        bar_column = BarColumn(bar_width=None)
        progress_text_column = TextColumn(
            "{task.completed} of {task.total} roots found"
        )
        progress = Progress(text_column, bar_column, progress_text_column, expand=True)
        task = progress.add_task("Rootfinding")
        progress.start()

        try:
            for roots, multiplicities, _, num_remaining_roots in root_finder:
                num_found_roots = sum(multiplicities)
                total_roots = num_found_roots + num_remaining_roots
                progress.update(task, completed=num_found_roots, total=total_roots)
        finally:
            progress.stop()

    else:
        for roots, multiplicities, _, _ in root_finder:
            pass

    return RootResult(roots, multiplicities, original_contour)



import os
import tempfile

import pytest
from numpy import cos, exp

from cxroots import Circle


@pytest.fixture()
def results_dir():
    with tempfile.TemporaryDirectory() as tmp:
        yield tmp


def test_save_plot(results_dir):
    file_path = os.path.join(results_dir, "circle.pdf")
    Circle(1j, 2).show(save_file=file_path)
    assert os.path.exists(file_path)


def test_save_animation(results_dir):
    file_path = os.path.join(results_dir, "circle.gif")
    Circle(1j, 2).demo_roots(f=lambda z: cos(exp(1j * z)), save_file=file_path)
    assert os.path.exists(file_path)


def test_save_roots(results_dir):
    file_path = os.path.join(results_dir, "roots.png")
    Circle(1j, 2).roots(f=lambda z: cos(exp(1j * z))).show(save_file=file_path)
    assert os.path.exists(file_path)



import functools
from collections.abc import Generator, Sequence
from typing import Optional, overload

import numpy as np
import numpy.typing as npt

from .contour_interface import ContourABC
from .paths import ComplexPath, ComplexPathType
from .root_counting import count_roots
from .root_finding import find_roots
from .root_finding_demo import demo_find_roots, demo_roots_animation
from .root_result import RootResult
from .types import AnalyticFunc
from .util import remove_para


class Contour(ContourABC):
    """
    A base class for contours in the complex plane.

    Attributes
    ----------
    central_point : complex
        The point at the center of the contour.
    area : float
        The surface area of the contour.
    """

    # Should be set in subclass
    axis_names = ()

    def __init__(self, segments: list[ComplexPathType]):
        self.segments = segments

        # A contour created by the subdvision method will have this attribute set to
        # the axis along which the line subdividing the parent contour was a constant.
        # This is done in order to implement the "alternating" subdivision method
        self._created_by_subdivision_axis: str | None = None
        # _parent and _children are set in subdivision method
        self._parent: Contour | None = None
        self._children: Sequence[Contour] | None = None

    @overload
    def __call__(self, t: float) -> complex: ...

    @overload
    def __call__(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def __call__(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        r"""
        The point on the contour corresponding the value of the
        parameter t.

        Parameters
        ----------
        t : float
            A real number :math:`0\leq t \leq 1` which parameterises
            the contour.

        Returns
        -------
        complex
            A point on the contour.

        Example
        -------

        >>> from cxroots import Circle
        >>> c = Circle(0,1) # Circle |z|=1 parameterised by e^{it}
        >>> c(0.25)
        (6.123233995736766e-17+1j)
        >>> c(0) == c(1)
        True
        """
        t = np.array(t, dtype=np.float_)
        num_segments = len(self.segments)
        segment_index = np.array(num_segments * t, dtype=int)
        segment_index = np.mod(segment_index, num_segments)

        if hasattr(segment_index, "__iter__"):
            return np.array(
                [
                    self.segments[i](num_segments * t[ti] % 1)
                    for ti, i in enumerate(segment_index)
                ],
                dtype=complex,
            )
        else:
            return self.segments[segment_index](num_segments * t % 1)

    def trap_product(self, *args, **integration_kwargs) -> complex:
        r"""
        Use Romberg integration to estimate the symmetric bilinear form used in
        (1.12) of [KB]_ using 2**k+1 samples

        .. math::

            <\phi,\psi> = \frac{1}{2\pi i} \oint_C \phi(z)\psi(z)\frac{f'(z)}{f(z)} dz
        """
        return sum(s.trap_product(*args, **integration_kwargs) for s in self.segments)

    functools.update_wrapper(
        trap_product, ComplexPath.trap_product, assigned=["__doc__", "__annotations__"]
    )

    def integrate(self, f: AnalyticFunc, **integration_kwargs) -> complex:
        r"""
        Integrate the function f along the contour C

        .. math::

            \oint_C f(z) dz
        """
        return sum(
            segment.integrate(f, **integration_kwargs) for segment in self.segments
        )

    functools.update_wrapper(
        integrate, ComplexPath.integrate, assigned=["__doc__", "__annotations__"]
    )

    def plot(self, *args, **kwargs) -> None:
        self.size_plot()
        for segment in self.segments:
            segment.plot(*args, **kwargs)

    functools.update_wrapper(
        plot, ComplexPath.plot, assigned=["__doc__", "__annotations__"]
    )

    def size_plot(self) -> None:
        """
        Adjust the plot axes limits to nicely frame the contour. Called as part of
        :meth:`~cxroots.contour.Contour.plot`
        """
        import matplotlib.pyplot as plt

        t = np.linspace(0, 1, 1000)
        z = self(t)
        xpad = (max(np.real(z)) - min(np.real(z))) * 0.1
        ypad = (max(np.imag(z)) - min(np.imag(z))) * 0.1

        xmin = min(np.real(z)) - xpad
        xmax = max(np.real(z)) + xpad
        ymin = min(np.imag(z)) - ypad
        ymax = max(np.imag(z)) + ypad
        plt.xlim([xmin, xmax])
        plt.ylim([ymin, ymax])

    def show(self, save_file: str | None = None, **plot_kwargs) -> None:
        """
        Shows the contour as a 2D plot in the complex plane.  Requires
        Matplotlib.

        Parameters
        ----------
        save_file : str (optional)
            If given then the plot will be saved to disk with name
            'save_file'.  If save_file=None the plot is shown on-screen.
        **plot_kwargs
            Key word arguments are as in :meth:`~cxroots.contour.Contour.plot`.
        """
        import matplotlib.pyplot as plt

        self.plot(**plot_kwargs)

        if save_file is not None:
            plt.savefig(save_file, bbox_inches="tight")
            plt.close()
        else:
            plt.show()

    def subdivide(self, axis, division_factor: float) -> tuple["Contour", ...]:
        """
        Subdivide the contour
        """
        raise NotImplementedError("subdivide must be implemented in a subclass")

    @property
    def parent(self) -> Optional["Contour"]:
        return self._parent

    @property
    def children(self) -> Sequence["Contour"] | None:
        return self._children

    def subdivisions(
        self, axis: str = "alternating"
    ) -> Generator[tuple["Contour", ...], None, None]:
        """
        A generator for possible subdivisions of the contour.

        Parameters
        ----------
        axis : str, 'alternating' or any element of self.axis_names.
            The axis along which the line subdividing the contour is a
            constant (eg. subdividing a circle along the radial axis
            will give an outer annulus and an inner circle).  If
            alternating then the dividing axis will always be different
            to the dividing axis used to create the contour which is now
            being divided.

        Yields
        ------
        tuple
            A tuple with two contours which subdivide the original
            contour.
        """
        if axis == "alternating":
            if self._created_by_subdivision_axis is None:
                axis_index = 0
            else:
                axis_index = (
                    self.axis_names.index(self._created_by_subdivision_axis) + 1
                ) % len(self.axis_names)

            axis = self.axis_names[axis_index]

        for division_factor in division_factor_gen():
            yield self.subdivide(axis, division_factor)

    def distance(self, z: complex) -> float:
        """
        The distance from the point z in the complex plane to the
        nearest point on the contour.

        Parameters
        ----------
        z : complex
            The point from which to measure the distance to the closest
            point on the contour to z.

        Returns
        -------
        float
            The distance from z to the point on the contour which is
            closest to z.
        """
        return min(segment.distance(z) for segment in self.segments)

    @remove_para("C")
    @functools.wraps(count_roots)
    def count_roots(
        self, f: AnalyticFunc, df: AnalyticFunc | None = None, **kwargs
    ) -> int:
        return count_roots(self, f, df, **kwargs)

    @remove_para("original_contour")
    @functools.wraps(find_roots)
    def roots(
        self, f: AnalyticFunc, df: AnalyticFunc | None = None, **kwargs
    ) -> RootResult:
        return find_roots(self, f, df, **kwargs)

    @remove_para("C")
    @functools.wraps(demo_find_roots)
    def demo_roots(self, *args, **kwargs) -> None:
        return demo_find_roots(self, *args, **kwargs)

    @remove_para("C")
    @functools.wraps(demo_roots_animation)
    def demo_roots_animation(self, *args, **kwargs):
        return demo_roots_animation(self, *args, **kwargs)


def division_factor_gen() -> Generator[float, None, None]:
    """A generator for division_factors."""
    yield 0.3  # being off-center is a better first choice for certain problems

    x = 0.5
    yield x
    for diff in np.linspace(0, 0.5, int(1 + 10 / 2.0))[1:-1]:
        yield x + diff
        yield x - diff



import logging
import warnings
from collections.abc import Callable
from math import inf, pi
from typing import overload

import numpy as np
import numpy.typing as npt

from .contour_interface import ContourABC
from .derivative import central_diff
from .types import AnalyticFunc, ComplexScalarOrArray, IntegrationMethod, ScalarOrArray

RombCallback = Callable[[complex, float | None, int], bool | None]


def prod(
    C: ContourABC,  # noqa: N803
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    phi: AnalyticFunc | None = None,
    psi: AnalyticFunc | None = None,
    abs_tol: float = 1.49e-08,
    rel_tol: float = 1.49e-08,
    div_min: int = 3,
    div_max: int = 15,
    int_method: IntegrationMethod = "quad",
    integer_tol: float = inf,
    callback: RombCallback | None = None,
) -> complex:
    r"""
    Compute the symmetric bilinear form used in (1.12) of [KB]_.

    .. math::

        <\phi,\psi> = \frac{1}{2\pi i} \oint_C \phi(z) \psi(z) \frac{f'(z)}{f(z)} dz.

    Parameters
    ----------
    C : :class:`Contour <cxroots.contour.Contour>`
        A contour in the complex plane for.  No roots or poles of f
        should lie on C.
    f : function
        Function of a single variable f(x)
    df : function, optional
        Function of a single variable, df(x), providing the derivative
        of the function f(x) at the point x.  If not provided then df is
        approximated using a finite difference method.
    phi : function, optional
        Function of a single variable phi(x).  If not provided then
        phi(z)=1.
    psi : function, optional
        Function of a single variable psi(x).  If not provided then
        psi(z)=1.
    abs_tol : float, optional
        Absolute error tolerance for integration.
    rel_tol : float, optional
        Relative error tolerance for integration.
    div_min : int, optional
        Only used if int_method='romb'. Minimum number of divisions before
        the Romberg integration routine is allowed to exit.
    div_max : int, optional
        Only used if int_method='romb'.  The maximum number of divisions
        before the Romberg integration routine of a path exits.
    int_method : {'quad', 'romb'}, optional
        If 'quad' then scipy.integrate.quad is used to perform the
        integral.  If 'romb' then Romberg integraion, using
        scipy.integrate.romb, is performed instead.
    integer_tol : float, optional
        Only used when int_method is 'romb'.  The integration routine will
        not exit unless the result is within integer_tol of an integer.
        This is useful when computing the number of roots in a contour,
        which must be an integer.  By default integer_tol is inf.
    callback : function, optional
        Only used when int_method is 'romb'.  A function that at each
        step in the iteration is passed the current approximation for
        the integral, the estimated error of that approximation and the
        number of iterations.  If the return of callback evaluates to
        True then the integration will end.

    Returns
    -------
    complex
        The value of the integral <phi, psi>.
    float
        An estimate of the error for the integration.

    References
    ----------
    .. [KB] "Computing the zeros of analytic functions" by Peter Kravanja,
        Marc Van Barel, Springer 2000
    """
    if int_method == "romb":
        return _romb_prod(
            C,
            f,
            df,
            phi,
            psi,
            abs_tol,
            rel_tol,
            div_min,
            div_max,
            integer_tol,
            callback,
        )
    elif int_method == "quad":
        return _quad_prod(C, f, df, phi, psi, abs_tol, rel_tol)
    else:
        raise ValueError("int_method must be either 'romb' or 'quad'")


def _romb_prod(
    C: ContourABC,  # noqa: N803
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    phi: AnalyticFunc | None = None,
    psi: AnalyticFunc | None = None,
    abs_tol: float = 1.49e-08,
    rel_tol: float = 1.49e-08,
    div_min: int = 3,
    div_max: int = 15,
    integer_tol: float = inf,
    callback: RombCallback | None = None,
) -> complex:
    logger = logging.getLogger(__name__)
    k = 0
    I = []  # List of approximations to the integral # noqa: E741 N806

    while k < div_max and (
        len(I) < div_min
        or (abs(I[-2] - I[-1]) > abs_tol and abs(I[-2] - I[-1]) > rel_tol * abs(I[-1]))
        or (abs(I[-3] - I[-2]) > abs_tol and abs(I[-3] - I[-2]) > rel_tol * abs(I[-2]))
        or abs(int(round(I[-1].real)) - I[-1].real) > integer_tol
        or abs(I[-1].imag) > integer_tol
    ):
        k += 1
        integral = C.trap_product(k, f, df, phi, psi)
        I.append(integral)
        if k > 1:
            logger.debug(f"Iteration={k}, integral={I[-1]}, err={I[-2] - I[-1]}")
        else:
            logger.debug(f"Iteration={k}, integral={I[-1]}")

        if callback is not None:
            err = abs(I[-2] - I[-1]) if k > 1 else None
            if callback(I[-1], err, k):
                break

    return I[-1]


def _quad_prod(
    C: ContourABC,  # noqa: N803
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    phi: AnalyticFunc | None = None,
    psi: AnalyticFunc | None = None,
    abs_tol: float = 1.49e-08,
    rel_tol: float = 1.49e-08,
) -> complex:
    if df is None:
        df = central_diff(f)

    def one(z: ScalarOrArray) -> int:
        return 1

    if phi is None:
        phi = one
    if psi is None:
        psi = one

    @overload
    def integrand_func(z: complex | float) -> complex: ...

    @overload
    def integrand_func(
        z: npt.NDArray[np.complex_] | npt.NDArray[np.float_],
    ) -> npt.NDArray[np.complex_] | complex: ...

    def integrand_func(z: ScalarOrArray) -> ComplexScalarOrArray:
        return phi(z) * psi(z) * (df(z) / f(z)) / (2j * pi)

    return C.integrate(
        integrand_func, abs_tol=abs_tol, rel_tol=rel_tol, int_method="quad"
    )


class RootError(RuntimeError):
    pass


def count_roots(
    C: ContourABC,  # noqa: N803
    f: AnalyticFunc,
    df: AnalyticFunc | None = None,
    int_abs_tol: float = 0.07,
    integer_tol: float = 0.1,
    div_min: int = 3,
    div_max: int = 15,
    int_method: IntegrationMethod = "quad",
) -> int:
    r"""
    For a function of one complex variable, f(z), which is analytic in
    and within the contour C, return the number of zeros (counting
    multiplicities) within the contour, N, using Cauchy's argument
    principle,

    .. math::

        N = \frac{1}{2i\pi} \oint_C \frac{f'(z)}{f(z)} dz.

    If df(z), the derivative of f(z), is provided then the above
    integral is computed directly.  Otherwise the derivative is
    approximated using a finite difference method.

    The number of roots is taken to be the closest integer to the
    computed value of the integral and the result is only accepted
    if the integral is within integer_tol of the closest integer.

    Parameters
    ----------
    C : :class:`Contour <cxroots.contour.Contour>`
        The contour which encloses the roots of f(z) that are to be
        counted.
    f : function
        Function of a single variable f(z).
    df : function, optional
        Function of a single complex variable, df(z), providing the
        derivative of the function f(z) at the point z.  If not
        provided, df will be approximated using a finite difference
        method.
    int_abs_tol : float, optional
        Required absolute error tolerance for the contour integration.
        Since the Cauchy integral must be an integer it is only
        necessary to distinguish which integer the integral is
        converging towards.  Therefore, int_abs_tol can be fairly large.
    integer_tol : float, optional
        The evaluation of the Cauchy integral will be accepted if its
        value is within integer_tol of the closest integer.
    div_min : int, optional
        Only used if int_method='romb'. Minimum number of divisions
        before the Romberg integration routine is allowed to exit.
    div_max : int, optional
        Only used if int_method='romb'. The maximum number of divisions
        before the Romberg integration routine of a path exits.
    int_method : {'quad', 'romb'}, optional
        If 'quad' then scipy.integrate.quad is used to perform the
        integral.  If 'romb' then Romberg integraion, using
        scipy.integrate.romb, is performed instead.

    Returns
    -------
    int
        The number of zeros of f (counting multiplicities) which lie
        within the contour C.
    """
    logger = logging.getLogger(__name__)
    logger.debug("Computing number of roots within " + str(C))

    with warnings.catch_warnings():
        # ignore warnings and catch if integral is NaN later
        warnings.simplefilter("ignore")
        integral = prod(
            C,
            f,
            df,
            abs_tol=int_abs_tol,
            rel_tol=0,
            div_min=div_min,
            div_max=div_max,
            int_method=int_method,
            integer_tol=integer_tol,
        )

    logger.debug(f"Integral for number of roots = {integral}")

    if np.isnan(integral):
        raise RootError(
            "Result of integral is an invalid value. "
            "Most likely because of a divide by zero error."
        )

    elif (
        abs(int(round(integral.real)) - integral.real) < integer_tol
        and abs(integral.imag) < integer_tol
    ):
        # integral is sufficiently close to an integer
        num_zeros = int(round(integral.real))
        logger.info(
            "Counted "
            + str(num_zeros)
            + " roots (including multiplicities) within "
            + str(C)
        )

        return num_zeros

    else:
        raise RootError("The number of enclosed roots has not converged to an integer")



import numpy as np
import pytest

from cxroots import AnnulusSector, Circle, Rectangle


def test_annulus_sector_contains():
    r0 = 8.938
    r1 = 9.625
    phi0 = 6.126
    phi1 = 6.519

    z = 9 - 1.04825594683e-18j
    contour = AnnulusSector(0, (r0, r1), (phi0, phi1))

    assert contour.contains(z)


def test_rect_contains():
    contour = Rectangle((-2355, -1860), (-8810, -8616))

    assert contour.contains(-2258 - 8694j)
    assert not contour.contains(-2258 - 8500j)


@pytest.mark.parametrize(
    "contour,t_arr,z_arr",
    [
        pytest.param(
            Rectangle((0, 2), (0, 1)),
            np.array([0, 0.125, 0.25, 0.5, 0.75, 1]),
            np.array([0, 1, 2, 2 + 1j, 1j, 0]),
            id="Rectangle",
        ),
        pytest.param(
            Circle(0, 1),
            np.array([0, 0.25]),
            np.array([1, 6.123233995736766e-17 + 1j]),
            id="Circle",
        ),
    ],
)
def test_contour_call(contour, t_arr, z_arr):
    # Test passing in the whole array
    assert np.all(contour(t_arr) == z_arr)

    # Test individual calls
    for t, z in zip(t_arr, z_arr):
        assert contour(t) == z



import numpy as np
import pytest
from numpy import cos, exp, sin

from cxroots import Rectangle


@pytest.mark.parametrize("use_df", [True, False])
def test_count_roots(use_df):
    """
    From "Locating all the Zeros of an Analytic Function in one Complex Variable"
    M.Dellnitz, O.Schutze, Q.Zheng, J. Compu. and App. Math. (2002), Vol.138, Issue 2

    There should be 424 roots inside this contour
    """
    contour = Rectangle([-20.3, 20.7], [-20.3, 20.7])

    def f(z):
        return z**50 + z**12 - 5 * sin(20 * z) * cos(12 * z) - 1

    if use_df:

        def df(z):
            return (
                50 * z**49
                + 12 * z**11
                + 60 * sin(12 * z) * sin(20 * z)
                - 100 * cos(12 * z) * cos(20 * z)
            )

    else:
        df = None

    assert contour.count_roots(f, df) == 424


@pytest.mark.xfail(reason="Possibly lack of high precision arithmetic")
def test_ring_oscillator():
    """Problem in section 4.3 of [DSZ]"""

    def A(z):  # noqa: N802
        t = 2.0
        return np.array(
            [
                [-0.0166689 - 2.12e-14 * z, 1 / 60.0 + 6e-16 * exp(-t * z) * z],
                [
                    0.0166659 + exp(-t * z) * (-0.000037485 + 6e-16 * z),
                    -0.0166667 - 6e-16 * z,
                ],
            ]
        )

    def dA(z):  # noqa: N802
        t = 2.0
        return np.array(
            [
                [
                    -2.12e-14 * np.ones_like(z),
                    6e-16 * exp(-t * z) - t * 6e-16 * exp(-t * z) * z,
                ],
                [
                    -t * exp(-t * z) * (-0.000037485 + 6e-16 * z) + exp(-t * z) * 6e-16,
                    -6e-16 * np.ones_like(z),
                ],
            ]
        )

    def f(z):
        A_val = np.rollaxis(A(z), -1, 0)  # noqa: N806
        return np.linalg.det(A_val)

    def df(z):
        A_val = A(z)  # noqa: N806
        dA_val = dA(z)  # noqa: N806
        return (
            dA_val[0, 0] * A_val[1, 1]
            + A_val[0, 0] * dA_val[1, 1]
            - dA_val[0, 1] * A_val[1, 0]
            - A_val[0, 1] * dA_val[1, 0]
        )

    box = Rectangle([-12, 0], [-40, 40])

    # XXX: No roots are recorded within the initial contour.
    #   Perhaps because the coefficents of z are very small?
    #   Perhaps need higher precision?
    assert box.count_roots(f, df) != 0
    assert box.count_roots(f) != 0

    # roots_fdf = findRoots(box, f, df)
    # roots_f = findRoots(box, f)

    # compare with fig 4 of [DSZ]
    # roots_fdf.show()



from typing import Literal, Protocol, Union, overload

import numpy as np
import numpy.typing as npt

IntegrationMethod = Literal["quad", "romb"]
Color = Union[str, tuple[float, float, float], tuple[float, float, float, float]]  # noqa: UP007
ScalarOrArray = Union[complex, float, npt.NDArray[np.complex_], npt.NDArray[np.float_]]  # noqa: UP007
ComplexScalarOrArray = Union[complex, npt.NDArray[np.complex_]]  # noqa: UP007


class AnalyticFunc(Protocol):
    @overload
    def __call__(self, z: complex | float) -> complex: ...

    @overload
    def __call__(
        self, z: npt.NDArray[np.complex_] | npt.NDArray[np.float_]
    ) -> ComplexScalarOrArray:
        # Note that the function may return a scalar in this case if, for example,
        # it's a constant function
        ...

    def __call__(self, z: ScalarOrArray) -> ComplexScalarOrArray: ...



from typing import Literal, overload

from numpy import pi

from ..contour import Contour
from ..paths import ComplexArc
from .annulus_sector import AnnulusSector


class Annulus(Contour):
    """
    An annulus in the complex plane with the outer circle positively oriented
    and the inner circle negatively oriented.

    Parameters
    ----------
    center : complex
            The center of the annulus in the complex plane.
    radii : tuple
            A tuple of length two of the form (inner_radius, outer_radius).

    Examples
    --------
    .. plot::
            :include-source:

            from cxroots import Annulus
            annulus = Annulus(center=0, radii=(0.5,0.75))
            annulus.show()
    """

    axis_names = ("r", "phi")

    def __init__(self, center: complex, radii: tuple[float, float]):
        self.center = center
        self.radii = radii

        segments = [
            ComplexArc(center, radii[1], 0, 2 * pi),
            ComplexArc(center, radii[0], 0, -2 * pi),
        ]
        super().__init__(segments)

    def __str__(self):
        return (
            f"Annulus: center={self.center.real:.3f}{self.center.imag:+.3f}i, "
            f"inner radius={self.radii[0]:.3f}, outer radius={self.radii[1]:.3f}"
        )

    @property
    def central_point(self) -> complex:
        # get a central point within the contour
        r = (self.radii[0] + self.radii[1]) / 2
        return r

    @property
    def area(self) -> float:
        return pi * (self.radii[1] ** 2 - self.radii[0] ** 2)

    def contains(self, z: complex) -> bool:
        """Returns True if the point z lies within the contour, False if otherwise"""
        return self.radii[0] < abs(z - self.center) < self.radii[1]

    @overload
    def subdivide(
        self, axis: Literal["r"], division_factor: float
    ) -> tuple["Annulus", "Annulus"]: ...

    @overload
    def subdivide(
        self, axis: Literal["phi"], division_factor: float
    ) -> tuple[AnnulusSector, AnnulusSector]: ...

    def subdivide(
        self, axis: Literal["r", "phi"], division_factor: float = 0.5
    ) -> tuple[AnnulusSector, AnnulusSector] | tuple["Annulus", "Annulus"]:
        """
        Subdivide the contour

        Parameters
        ----------
        axis : str, can be either 'r' or 'phi'
            The axis along which the line subdividing the contour is a constant.
        division_factor : float in range (0,1), optional
            Determines the point along 'axis' at which the line dividing the box is
            placed

        Returns
        -------
        boxes : list of contours
            Two annuluses if axis is 'r'.
            Two half-annuluses oriented according to division_factor if axis is 'phi'.
        """
        if axis == "r":
            midpoint = self.radii[0] + division_factor * (self.radii[1] - self.radii[0])
            boxes = (
                Annulus(self.center, (self.radii[0], midpoint)),
                Annulus(self.center, (midpoint, self.radii[1])),
            )

            boxes[0].segments[1] = self.segments[1]
            boxes[1].segments[0] = self.segments[0]
            boxes[0].segments[0]._reverse_path = boxes[1].segments[1]
            boxes[1].segments[1]._reverse_path = boxes[0].segments[0]

        elif axis == "phi":
            # Subdividing into two radial boxes rather than one to
            # ensure that an error is raised if one of the new paths
            # is too close to a root
            # XXX: introduce another parameter for phi1

            phi0 = 2 * pi * division_factor
            phi1 = phi0 + pi
            boxes = (
                AnnulusSector(self.center, self.radii, (phi0, phi1)),
                AnnulusSector(self.center, self.radii, (phi1, phi0)),
            )

            boxes[0].segments[0]._reverse_path = boxes[1].segments[2]
            boxes[1].segments[2]._reverse_path = boxes[0].segments[0]
            boxes[0].segments[2]._reverse_path = boxes[1].segments[0]
            boxes[1].segments[0]._reverse_path = boxes[0].segments[2]

        else:
            raise ValueError("axis must be 'r' or 'phi'")

        for box in boxes:
            box._created_by_subdivision_axis = axis
            box._parent = self
        self._children = boxes

        return boxes



import pytest
from numpy import cos, pi

from cxroots.iterative_methods import secant


def test_secant():
    """
    Example from Table 2.5 of "Numerical Analysis" by Richard L. Burden,
    J. Douglas Faires
    """

    def f(x):
        return cos(x) - x

    iterations = []

    def callback(x, dx, y, iteration):
        return iterations.append(x)

    secant(0.5, pi / 4, f, callback=callback)

    correct_iterations = [
        0.7363841388,
        0.7390581392,
        0.7390851493,
        0.7390851332,
        0.7390851332,
    ]

    assert iterations == pytest.approx(correct_iterations)



import unittest

from numpy import pi

from cxroots import Annulus, AnnulusSector, Circle, Rectangle
from cxroots.tests.approx_equal import roots_approx_equal


class TestRootfindingContours(unittest.TestCase):
    def setUp(self):
        self.roots = roots = [0, -1.234, 1 + 1j, 1 - 1j, 2.345]
        self.multiplicities = [1, 1, 1, 1, 1]
        self.f = (
            lambda z: (z - roots[0])
            * (z - roots[1])
            * (z - roots[2])
            * (z - roots[3])
            * (z - roots[4])
        )
        self.df = (
            lambda z: (z - roots[1]) * (z - roots[2]) * (z - roots[3]) * (z - roots[4])
            + (z - roots[0]) * (z - roots[2]) * (z - roots[3]) * (z - roots[4])
            + (z - roots[0]) * (z - roots[1]) * (z - roots[3]) * (z - roots[4])
            + (z - roots[0]) * (z - roots[1]) * (z - roots[2]) * (z - roots[4])
            + (z - roots[0]) * (z - roots[1]) * (z - roots[2]) * (z - roots[3])
        )

        self.Circle = Circle(0, 3)
        self.Rectangle = Rectangle([-2, 2], [-2, 2])
        self.halfAnnulus = AnnulusSector(0, [0.5, 3], [-pi / 2, pi / 2])
        self.Annulus = Annulus(0, [1, 2])

    def test_rootfinding_circle_fdf(self):
        roots_approx_equal(
            self.Circle.roots(self.f, self.df),
            (self.roots, self.multiplicities),
            decimal=7,
        )

    def test_rootfinding_circle_f(self):
        roots_approx_equal(
            self.Circle.roots(self.f, self.df),
            (self.roots, self.multiplicities),
            decimal=7,
        )

    def test_rootfinding_rectangle_fdf(self):
        roots_approx_equal(
            self.Rectangle.roots(self.f, self.df),
            (self.roots[:-1], self.multiplicities[:-1]),
            decimal=7,
        )

    def test_rootfinding_rectangle_f(self):
        roots_approx_equal(
            self.Rectangle.roots(self.f, self.df),
            (self.roots[:-1], self.multiplicities[:-1]),
            decimal=7,
        )

    def test_rootfinding_half_annulus_fdf(self):
        roots_approx_equal(
            self.halfAnnulus.roots(self.f, self.df),
            (self.roots[2:], self.multiplicities[2:]),
            decimal=7,
        )

    def test_rootfinding_half_annulus_f(self):
        roots_approx_equal(
            self.halfAnnulus.roots(self.f, self.df),
            (self.roots[2:], self.multiplicities[2:]),
            decimal=7,
        )

    def test_rootfinding_annulus_fdf(self):
        roots_approx_equal(
            self.Annulus.roots(self.f, self.df),
            (self.roots[1:-1], self.multiplicities[1:-1]),
            decimal=7,
        )

    def test_rootfinding_annulus_f(self):
        roots_approx_equal(
            self.Annulus.roots(self.f, self.df),
            (self.roots[1:-1], self.multiplicities[1:-1]),
            decimal=7,
        )



from cxroots import Circle


def test_progress():
    """Test progress bar (verbose=True) works without error"""
    contour = Circle(0, 3)

    def f(z):
        return (z + 1.2) ** 3 * (z - 2.5) ** 2 * (z + 1j)

    contour.roots(f, verbose=True)



from typing import Literal

from ..contour import Contour
from ..paths import ComplexLine


class Rectangle(Contour):
    """
    A positively oriented rectangle in the complex plane.

    Parameters
    ----------
    x_range : tuple
            Tuple of length two giving the range of the rectangle along the
            real axis.
    y_range : tuple
            Tuple of length two giving the range of the rectangle along the
            imaginary axis.

    Examples
    --------
    .. plot::
            :include-source:

            from cxroots import Rectangle
            rect = Rectangle(x_range=(-2, 2), y_range=(-1, 1))
            rect.show()
    """

    axis_names = ("x", "y")

    def __init__(self, x_range: tuple[float, float], y_range: tuple[float, float]):
        self.x_range = x_range
        self.y_range = y_range

        self.z1 = z1 = self.x_range[0] + 1j * self.y_range[0]
        self.z2 = z2 = self.x_range[1] + 1j * self.y_range[0]
        self.z3 = z3 = self.x_range[1] + 1j * self.y_range[1]
        self.z4 = z4 = self.x_range[0] + 1j * self.y_range[1]

        segments = [
            ComplexLine(z1, z2),
            ComplexLine(z2, z3),
            ComplexLine(z3, z4),
            ComplexLine(z4, z1),
        ]
        super().__init__(segments)

    def __str__(self):
        return (
            "Rectangle: vertices = "
            f"{self.z1.real:.3f}{self.z1.imag:+.3f}i, "
            f"{self.z2.real:.3f}{self.z2.imag:+.3f}i, "
            f"{self.z3.real:.3f}{self.z3.imag:+.3f}i, "
            f"{self.z4.real:.3f}{self.z4.imag:+.3f}i"
        )

    @property
    def central_point(self) -> complex:
        # get the central point within the contour
        x = (self.x_range[0] + self.x_range[1]) / 2
        y = (self.y_range[0] + self.y_range[1]) / 2
        return x + 1j * y

    @property
    def area(self) -> float:
        return (self.x_range[1] - self.x_range[0]) * (self.y_range[1] - self.y_range[0])

    def contains(self, z: complex) -> bool:
        """Returns True if the point z lies within the contour, False if otherwise"""
        return (
            self.x_range[0] < z.real < self.x_range[1]
            and self.y_range[0] < z.imag < self.y_range[1]
        )

    def subdivide(
        self, axis: Literal["x", "y"], division_factor: float = 0.5
    ) -> tuple["Rectangle", "Rectangle"]:
        """
        Subdivide the contour

        Parameters
        ----------
        axis : str, can be either 'x' or 'y'
            The axis along which the line subdividing the contour is a constant.
        division_factor : float in range (0,1), optional
            Determines the point along 'axis' at which the line dividing the contour
            is placed.

        Returns
        -------
        box1 : Rectangle
            If axis is 'x' then box1 has the same y_range and minimum value of x_range
            as the original Rectangle but the maximum x_range is determined by the
            division_factor.
            If axis is 'y' then box1 has the same x_range and minimum value of y_range
            as the original Rectangle but the maximum y_range is determined by the
            division_factor.
        box2 : Rectangle
            If axis is 'x' then box2 has the same y_range and maximum value of x_range
            as the original Rectangle but the minimum x_range is equal to the maximum
            x_range of box1.
            If axis is 'x' then box2 has the same x_range and maximum value of y_range
            as the original Rectangle but the minimum y_range is equal to the maximum
            y_range of box1.
        """
        if axis == "x":
            midpoint = self.x_range[0] + division_factor * (
                self.x_range[1] - self.x_range[0]
            )
            box1 = Rectangle((self.x_range[0], midpoint), self.y_range)
            box2 = Rectangle((midpoint, self.x_range[1]), self.y_range)

            box1.segments[3] = self.segments[3]
            box2.segments[1] = self.segments[1]
            box1.segments[1]._reverse_path = box2.segments[3]
            box2.segments[3]._reverse_path = box1.segments[1]

        elif axis == "y":
            midpoint = self.y_range[0] + division_factor * (
                self.y_range[1] - self.y_range[0]
            )
            box1 = Rectangle(self.x_range, (self.y_range[0], midpoint))
            box2 = Rectangle(self.x_range, (midpoint, self.y_range[1]))

            box1.segments[0] = self.segments[0]
            box2.segments[2] = self.segments[2]
            box1.segments[2]._reverse_path = box2.segments[0]
            box2.segments[0]._reverse_path = box1.segments[2]

        else:
            raise ValueError("axis must be 'x' or 'y'")

        for box in [box1, box2]:
            box._created_by_subdivision_axis = axis
            box._parent = self
        self._children = [box1, box2]

        return box1, box2



#!/usr/bin/env python
from setuptools import setup

packages = ["cxroots", "cxroots.tests", "cxroots.contours"]

# get the version, this will assign __version__
with open("cxroots/version.py") as f:
    exec(f.read())  # nosec

# read the README_pip.rst
try:
    with open("README.rst") as file:
        long_description = file.read()
except OSError:
    long_description = None

setup(
    name="cxroots",
    version=__version__,  # noqa
    description="Find all the roots (zeros) of a complex analytic function within a "
    "given contour in the complex plane.",
    long_description=long_description,
    author="Robert Parini",
    author_email="robert.parini@gmail.com",
    url="https://rparini.github.io/cxroots/",
    license="BSD-3-Clause",
    data_files=[("", ["LICENSE"])],
    packages=packages,
    package_data={"cxroots": ["py.typed"]},
    zip_safe=False,  # prevent cxroots from installing as a .egg zip file
    platforms=["all"],
    python_requires=">=3.10",
    setup_requires=["pytest-runner"],
    install_requires=[
        "numpy",
        "scipy",
        "numpydoc",
        "mpmath",
        "rich",
    ],
    tests_require=["pytest"],
    extras_require={"plot": ["matplotlib"]},
    keywords="roots zeros complex analytic functions",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



import logging
from collections.abc import Callable

from mpmath import mp, mpmathify
from mpmath.calculus.optimization import Muller
from numpy import inf

Callback = Callable[[complex, complex, complex, int], bool]
ScalarCxFunc = Callable[[complex | float], complex]


def iterate_to_root(
    x0: complex,
    f: ScalarCxFunc,
    df: ScalarCxFunc | None = None,
    step_tol: float = 1e-12,
    root_tol: float = 1e-12,
    max_iter: int = 20,
    refine_roots_beyond_tol: bool = False,
    callback: Callback | None = None,
) -> complex | None:
    """
    Starting with initial point x0 iterate to a root of f. This function is called
    during the rootfinding process to refine any roots found. If df is given then
    the Newton-Raphson method, :func:`~cxroots.iterative_methods.newton`, will be used,
    otherwise Muller's method, :func:`~cxroots.iterative_methods.muller`, will be used
    instead.

    Parameters
    ----------
    x0 : complex
        An initial point for the iteration.
    f : function
        Function of a single variable which we seek to find a root of.
    df : function, optional
        The derivative of f.
    step_tol: float, optional
        The routine ends if the step size, dx, between sucessive
        iterations satisfies abs(dx) < step_tol and refine_roots_beyond_tol is False.
    root_tol: float, optional
        A root, x, is only returned if abs(f(x)) < root_tol, otherwise None is returned
    max_iter : int, optional
        The routine ends after max_iter iterations.
    refine_roots_beyond_tol : bool, optional
        If True then the routine ends only once the error of the previous iteration,
        x0, was at least as good as the current iteration, x, in the sense that
        abs(f(x)) >= abs(f(x0)), and the previous iteration satisfied
        abs(dx0) < step_tol. In this case the previous iteration is returned as the
        approximation of the root, provided that it satisfies abs(f(x)) < root_tol
    callback : function, optional
        After each iteration callback(x, dx, f(x), iteration) will be
        called where 'x' is the current iteration of the estimated root,
        'dx' is the step size between the previous and current 'x' and
        'iteration' the number of iterations that have been taken.  If
        the callback function evaluates to True then the routine will end.

    Returns
    -------
    complex
        An approximation for a root of f. If the rootfinding was
        unsucessful then None will be returned instead.
    """
    logger = logging.getLogger(__name__)
    logger.debug("Refining root: " + str(x0))

    if df is not None:
        try:
            root, err = newton(
                x0, f, df, step_tol, 0, max_iter, refine_roots_beyond_tol, callback
            )
        except (RuntimeError, OverflowError):
            return None
    else:
        # Muller's method:
        x1, x2, x3 = x0, x0 * (1 + 1e-8) + 1e-8j, x0 * (1 - 1e-8) - 1e-8j
        root, err = muller(
            x1,
            x2,
            x3,
            f,
            step_tol,
            0,
            max_iter,
            refine_roots_beyond_tol,
            callback,
        )

    if err < root_tol:
        return root


def muller(
    x1: complex,
    x2: complex,
    x3: complex,
    f: ScalarCxFunc,
    step_tol: float = 1e-12,
    root_tol: float = 0,
    max_iter: int = 20,
    refine_roots_beyond_tol: bool = False,
    callback: Callback | None = None,
) -> tuple[complex, float]:
    """
    A wrapper for mpmath's implementation of Muller's method.

    Parameters
    ----------
    x1 : float or complex
        An initial point for iteration, should be close to a root of f.
    x2 : float or complex
        An initial point for iteration, should be close to a root of f.
        Should not equal x1.
    x3 : float or complex
        An initial point for iteration, should be close to a root of f.
        Should not equal x1 or x2.
    f : function
        Function of a single variable which we seek to find a root of.
    step_tol: float, optional
        The routine ends if the step size, dx, between sucessive
        iterations satisfies abs(dx) < step_tol and refine_roots_beyond_tol is False.
    root_tol: float, optional
        The routine ends if abs(f(x)) < root_tol and refine_roots_beyond_tol is False.
    max_iter : int, optional
        The routine ends after max_iter iterations.
    refine_roots_beyond_tol : bool, optional
        If True then routine ends if the error of the previous iteration,
        x0, was at least as good as the current iteration, x, in the
        sense that abs(f(x)) >= abs(f(x0)) and the previous iteration
        satisfied either abs(dx0) < step_tol or abs(f(x0)) < root_tol.  In
        this case the previous iteration is returned as the approximation
        of the root.
    callback : function, optional
        After each iteration callback(x, dx, f(x), iteration) will be
        called where 'x' is the current iteration of the estimated root,
        'dx' is the step size between the previous and current 'x' and
        'iteration' the number of iterations that have been taken.  If
        the callback function evaluates to True then the routine will end.

    Returns
    -------
    complex
        The approximation to a root of f.
    float
        abs(f(x)) where x is the final approximation for the root of f.
    """
    logger = logging.getLogger(__name__)

    # mpmath insists on functions accepting mpc
    def f_mpmath(z):
        return mpmathify(f(complex(z)))

    mull = Muller(mp, f_mpmath, (x1, x2, x3), verbose=False)
    x0 = x3

    x, err = x0, abs(f(x0))
    err0, dx0 = inf, inf
    try:
        for iteration, (x, dx) in enumerate(mull):
            y = f_mpmath(x)
            err = abs(y)
            logger.debug(
                str(iteration)
                + " x="
                + str(x)
                + " |f(x)|="
                + str(err)
                + " dx="
                + str(dx)
            )

            if callback is not None and callback(x, dx, y, iteration + 1):
                break

            if (
                not refine_roots_beyond_tol
                and (abs(dx) < step_tol or err < root_tol)
                or iteration > max_iter
            ):
                break

            if (
                refine_roots_beyond_tol
                and (abs(dx0) < step_tol or err0 < root_tol)
                and err >= err0
            ):
                # The previous iteration was a better appproximation the current one so
                # assume that that was as close to the root as we are going to get.
                x, err = x0, err0
                break

            x0 = x

            if refine_roots_beyond_tol:
                # record previous error for comparison
                dx0, err0 = dx, err

    except ZeroDivisionError:
        # ZeroDivisionError comes up if the error is evaluated to be zero
        pass

    # cast mpc and mpf back to regular complex and float
    logger.debug(
        "Final approximation: x=" + str(complex(x)) + " |f(x)|=" + str(float(err))
    )
    return complex(x), float(err)


def newton(
    x0: complex,
    f: ScalarCxFunc,
    df: ScalarCxFunc,
    step_tol: float = 1e-12,
    root_tol: float = 0,
    max_iter: int = 20,
    refine_roots_beyond_tol: bool = False,
    callback: Callback | None = None,
) -> tuple[complex, float]:
    """
    Find an approximation to a point xf such that f(xf)=0 for a
    scalar function f using Newton-Raphson iteration starting at
    the point x0.

    Parameters
    ----------
    x0 : float or complex
        Initial point for Newton iteration, should be as close as
        possible to a root of f
    f : function
        Function of a single variable which we seek to find a root of.
    df : function
        Function of a single variable, df(x), providing the
        derivative of the function f(x) at the point x
    step_tol: float, optional
        The routine ends if the step size, dx, between sucessive
        iterations satisfies abs(dx) < step_tol and refine_roots_beyond_tol is False.
    root_tol: float, optional
        The routine ends if abs(f(x)) < root_tol and refine_roots_beyond_tol is False.
    max_iter : int, optional
        The routine ends after max_iter iterations.
    refine_roots_beyond_tol : bool, optional
        If True then routine ends if the error of the previous iteration,
        x0, was at least as good as the current iteration, x, in the
        sense that abs(f(x)) >= abs(f(x0)) and the previous iteration
        satisfied either abs(dx0) < step_tol or abs(f(x0)) < root_tol.  In
        this case the previous iteration is returned as the approximation
        of the root.
    callback : function, optional
        After each iteration callback(x, dx, f(x), iteration) will be
        called where 'x' is the current iteration of the estimated root,
        'dx' is the step size between the previous and current 'x' and
        'iteration' the number of iterations that have been taken.  If
        the callback function evaluates to True then the routine will end.

    Returns
    -------
    complex
        The approximation to a root of f.
    float
        abs(f(x)) where x is the final approximation for the root of f.
    """
    logger = logging.getLogger(__name__)
    x, y = x0, f(x0)
    dx0, y0 = inf, y

    for iteration in range(max_iter):
        dx = -y / df(x)
        x += dx
        y = f(x)

        logger.debug("x=" + str(x) + " f(x)=" + str(y) + " dx=" + str(dx))

        if callback is not None and callback(x, dx, y, iteration + 1):
            break

        if not refine_roots_beyond_tol and (abs(dx) < step_tol or abs(y) < root_tol):
            break

        if (
            refine_roots_beyond_tol
            and (abs(dx0) < step_tol or abs(y0) < root_tol)
            and abs(y) >= abs(y0)
        ):
            break

        if refine_roots_beyond_tol:
            # store previous dx and y
            dx0, y0 = dx, y

    logger.debug("Final approximation: x=" + str(x) + " |f(x)|=" + str(abs(y)))
    return x, abs(y)


def secant(
    x1: complex,
    x2: complex,
    f: ScalarCxFunc,
    step_tol: float = 1e-12,
    root_tol: float = 0,
    max_iter: int = 30,
    callback: Callback | None = None,
) -> tuple[complex, float]:
    """
    Find an approximation to a point xf such that f(xf)=0 for a
    scalar function f using the secant method.  The method requires
    two initial points x1 and x2, ideally close to a root,
    and proceeds iteratively.

    Parameters
    ----------
    x1 : float or complex
        An initial point for iteration, should be close to a
        root of f.
    x2 : float or complex
        An initial point for iteration, should be close to a
        root of f.  Should not equal x1.
    f : function
        Function of a single variable which we seek to find a root of.
    step_tol: float, optional
        The routine ends if the step size, dx, between sucessive
        iterations satisfies abs(dx) < step_tol and refine_roots_beyond_tol is False.
    root_tol: float, optional
        The routine ends if abs(f(x)) < root_tol and refine_roots_beyond_tol is False.
    max_iter : int, optional
        The routine ends after max_iter iterations.
    callback : function, optional
        After each iteration callback(x, dx, f(x), iteration) will be
        called where 'x' is the current iteration of the estimated root,
        'dx' is the step size between the previous and current 'x' and
        'iteration' the number of iterations that have been taken.  If
        the callback function evaluates to True then the routine will end.

    Returns
    -------
    complex
        The approximation to a root of f.
    float
        abs(f(x)) where x is the final approximation for the root of f.
    """
    # As in "Numerical Recipies 3rd Edition" pick the bound with the
    # smallest function value as the most recent guess
    y1, y2 = f(x1), f(x2)
    if abs(y1) < abs(y2):
        x1, x2 = x2, x1
        y1, y2 = y2, y1

    for iteration in range(max_iter):
        dx = -(x2 - x1) * y2 / (y2 - y1)
        x1, x2 = x2, x2 + dx
        y1, y2 = y2, f(x2)

        if callback is not None and callback(x2, dx, y2, iteration + 1):
            break

        if abs(dx) < step_tol or abs(y2) < root_tol:
            break

    return x2, abs(y2)



from typing import Literal

from numpy import pi

from ..contour import Contour
from ..paths import ComplexArc
from .annulus import Annulus


class Circle(Contour):
    """
    A positively oriented circle in the complex plane.

    Parameters
    ----------
    center : complex
        The center of the circle.
    radius : float
        The radius of the circle.

    Examples
    --------
    .. plot::
        :include-source:

        from cxroots import Circle
        circle = Circle(center=1, radius=0.5)
        circle.show()
    """

    axis_names = tuple("r")

    def __init__(self, center: complex, radius: float):
        self.center = center
        self.radius = radius

        segments = [ComplexArc(center, radius, 0, 2 * pi)]
        super().__init__(segments)

    def __str__(self):
        return (
            f"Circle: center={self.center.real:.3f}{self.center.imag:+.3f}i, "
            f"radius={self.radius:.3f}"
        )

    def contains(self, z: complex) -> bool:
        """Returns True if the point z lies within the contour, False if otherwise"""
        return abs(z - self.center) < self.radius

    @property
    def central_point(self) -> complex:
        return self.center

    @property
    def area(self) -> float:
        return pi * self.radius**2

    def subdivide(
        self, axis: Literal["r"] = "r", division_factor: float = 0.5
    ) -> tuple[Annulus, "Circle"]:
        """
        Subdivide the contour

        Parameters
        ----------
        axis : str, can only be 'r'
            The axis along which the line subdividing the contour is a constant.
        division_factor : float in range (0,1), optional
            Determines the point along 'axis' at which the line dividing the box is
            placed

        Returns
        -------
        box1 : Annulus
            With inner radius determined by the division_factor and outer radius equal
            to that of the original circle
        box2 : Circle
            With radius equal to the inner radius of box1
        """
        if axis == "r":
            box1 = Annulus(self.center, (self.radius * division_factor, self.radius))
            box2 = Circle(self.center, self.radius * division_factor)
            box1.segments[0] = self.segments[0]
            box1.segments[1]._reverse_path = box2.segments[0]
            box2.segments[0]._reverse_path = box1.segments[1]

        else:
            raise ValueError("axis must be 'r'")

        for box in [box1, box2]:
            box._created_by_subdivision_axis = axis
            box._parent = self
        self._children = [box1, box2]

        return box1, box2



import pytest
from numpy import cos, pi

from cxroots.iterative_methods import muller


def test_muller():
    def f(x):
        return cos(x) - x**2 + 1j * x**3

    x, _ = muller(0.5, pi / 4, 0.6, f)
    root = 0.7296100078977741539157356847 + 0.1570923181734581909733787621j
    assert x == pytest.approx(root)



import pytest
from numpy import cos, pi, sin

from cxroots.iterative_methods import newton


@pytest.mark.parametrize("refine_roots_beyond_tol", [True, False])
def test_newton(refine_roots_beyond_tol):
    # result from online calculator: http://keisan.casio.com/exec/system/1244946907

    def f(x):
        return x - cos(x)

    def df(x):
        return 1 + sin(x)

    iterations = []

    def callback(x, dx, y, iteration):
        iterations.append((x, dx, y))

    x0 = pi / 4
    step_tol = 1e-10
    newton(
        x0,
        f,
        df,
        step_tol=step_tol,
        callback=callback,
        refine_roots_beyond_tol=refine_roots_beyond_tol,
    )

    correct_x = (
        0.7395361335152383009399981259049958,
        0.7390851781060101829533431920344887,
        0.7390851332151610866197574459477856,
        0.7390851332151606416553120876739171,
        0.7390851332151606416553120876738734,
        0.7390851332151606416553120876738734,
    )

    correct_iterations = [
        (x, x - x0 if i == 0 else x - correct_x[i - 1], f(x))
        for i, x in enumerate(correct_x)
    ]

    for i, iteration in enumerate(iterations):
        assert iteration == pytest.approx(correct_iterations[i])

    # dx should be less than step_tol
    assert abs(iterations[-1][1]) < step_tol

    if not refine_roots_beyond_tol:
        # we should have stopped as soon as step_tol was reached to the iteration
        # before the last should not satisfy the step_tol
        assert abs(iterations[-2][1]) > step_tol



import functools
from math import pi
from typing import TypeVar, overload

import numpy as np
import numpy.typing as npt
import scipy.integrate

from .types import AnalyticFunc, Color, IntegrationMethod

ComplexPathType = TypeVar("ComplexPathType", bound="ComplexPath")


class ComplexPath:
    """A base class for paths in the complex plane."""

    def __init__(self):
        self._integral_cache = {}
        self._trap_cache = {}

        # If the path is created during subvision then this will be set to the path
        # that is oppositely oriented to this path. This is done so that we can look
        # up the _integral_cache for the reverse path
        self._reverse_path: ComplexPath | None = None

    @overload
    def __call__(self, t: float) -> complex: ...

    @overload
    def __call__(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def __call__(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        r"""
        The parameterization of the path in the varaible :math:`t\in[0,1]`.

        Parameters
        ----------
        t : float
            A real number :math:`0\leq t \leq 1`.

        Returns
        -------
        complex
            A point on the path in the complex plane.
        """
        raise NotImplementedError("__call__ must be implemented in a subclass")

    @overload
    def dzdt(self, t: float) -> complex: ...

    @overload
    def dzdt(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def dzdt(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        """
        The derivative of the parameterised curve in the complex plane, z, with
        respect to the parameterization parameter, t.
        """
        raise NotImplementedError("dzdt must be implemented in a subclass")

    def distance(self, z: complex) -> float:
        """
        Distance from the point z to the closest point on the line.
        """
        raise NotImplementedError("distance must be implemented in a subclass")

    def trap_values(
        self,
        f: AnalyticFunc,
        k: int,
        use_cache: bool = True,
    ) -> npt.NDArray[np.complex128]:
        """
        Compute or retrieve (if cached) the values of the functions f
        at :math:`2^k+1` points along the contour which are evenly
        spaced with respect to the parameterisation of the contour.

        Parameters
        ----------
        f : function
            A function of a single complex variable.
        k : int
            Defines the number of points along the curve that f is to be
            evaluated at as :math:`2^k+1`.
        use_cache : bool, optional
            If True then use, if available, the results of any previous
            calls to this function for the same f and save any new
            results so that they can be reused later.

        Returns
        -------
        :class:`numpy.ndarray`
            The values of f at :math:`2^k+1` points along the contour
            which are evenly spaced with respect to the parameterisation
            of the contour.
        """
        if f in self._trap_cache and use_cache:
            vals = self._trap_cache[f]
            vals_k = int(np.log2(len(vals) - 1))

            if vals_k == k:
                return vals
            elif vals_k > k:
                return vals[:: 2 ** (vals_k - k)]
            else:
                t = np.linspace(0, 1, 2**k + 1)
                vals = np.empty(2**k + 1, dtype=np.complex128)
                vals.fill(np.nan)
                vals[:: 2 ** (k - vals_k)] = self._trap_cache[f]
                vals[np.isnan(vals)] = f(self(t[np.isnan(vals)]))

                # cache values
                self._trap_cache[f] = vals
                return vals

        else:
            t = np.linspace(0, 1, 2**k + 1)
            vals = f(self(t))
            if not isinstance(vals, np.ndarray):
                # Handle case when f does not return a vector, perhaps
                # because the function is actually a constant
                vals = vals * np.ones(len(t), dtype=np.complex128)

            if use_cache:
                self._trap_cache[f] = vals
            return vals

    def trap_product(
        self,
        k: int,
        f: AnalyticFunc,
        df: AnalyticFunc | None = None,
        phi: AnalyticFunc | None = None,
        psi: AnalyticFunc | None = None,
    ) -> complex:
        r"""
        Use Romberg integration to estimate the symmetric bilinear form used in
        (1.12) of [KB]_ using 2**k+1 samples

        .. math::

            <\phi,\psi> = \frac{1}{2\pi i} \oint_C \phi(z)\psi(z)\frac{f'(z)}{f(z)} dz
        """
        # compute/retrieve function evaluations
        f_val = self.trap_values(f, k)
        t = np.linspace(0, 1, 2**k + 1)
        dt = t[1] - t[0]

        if df is None:
            # approximate df/dz with finite difference
            dfdt = np.gradient(f_val, dt)
            df_val = dfdt / self.dzdt(t)
        else:
            df_val = self.trap_values(df, k)

        segment_integrand = df_val / f_val * self.dzdt(t)
        if phi is not None:
            segment_integrand = self.trap_values(phi, k) * segment_integrand
        if psi is not None:
            segment_integrand = self.trap_values(psi, k) * segment_integrand

        return scipy.integrate.romb(segment_integrand, dx=dt, axis=-1) / (2j * pi)

    def plot(
        self,
        num_points: int = 100,
        linecolor: Color = "C0",
        linestyle: str = "-",
    ) -> None:
        """
        Uses matplotlib to plot, but not show, the path as a 2D plot in
        the Complex plane.

        Parameters
        ----------
        num_points : int, optional
            The number of points to use when plotting the path.
        linecolor : optional
            The colour of the plotted path, passed to the
            :func:`matplotlib.pyplot.plot` function as the keyword
            argument of 'color'.  See the matplotlib tutorial on specifying
            `colours <https://matplotlib.org/stable/tutorials/colors/colors#>`_.
        linestyle : str, optional
            The line style of the plotted path, passed to the
            :func:`matplotlib.pyplot.plot` function as the keyword
            argument of 'linestyle'.  The default corresponds to a solid
            line.  See :meth:`matplotlib.lines.Line2D.set_linestyle` for
            other acceptable arguments.
        """
        import matplotlib.pyplot as plt

        t = np.linspace(0, 1, num_points)
        path = self(t)
        plt.plot(path.real, path.imag, color=linecolor, linestyle=linestyle)
        plt.xlabel("Re[$z$]", size=16)
        plt.ylabel("Im[$z$]", size=16)
        plt.gca().set_aspect(1)

        # add arrow to indicate direction of path
        arrow_start = self(0.5)
        arrow_end = self(0.51)

        plt.annotate(
            "",
            (arrow_end.real, arrow_end.imag),
            (arrow_start.real, arrow_start.imag),
            arrowprops=dict(arrowstyle="->", fc=linecolor, ec=linecolor),
        )

    def integrate(
        self,
        f: AnalyticFunc,
        abs_tol: float = 1.49e-08,
        rel_tol: float = 1.49e-08,
        div_max: int = 15,
        int_method: IntegrationMethod = "quad",
    ) -> complex:
        r"""
        Integrate the function f along the path.

        .. math::

            \oint_C f(z) dz

        The value of the integral is cached and will be reused if the method
        is called with same arguments.

        Parameters
        ----------
        f : function
            A function of a single complex variable.
        abs_tol : float, optional
            The absolute tolerance for the integration.
        rel_tol : float, optional
            The realative tolerance for the integration.
        div_max : int, optional
            If the Romberg integration method is used then div_max is the
            maximum number of divisions before the Romberg integration
            routine of a path exits.
        int_method : {'quad', 'romb'}, optional
            If 'quad' then :func:`scipy.integrate.quad` is used to
            compute the integral.  If 'romb' then Romberg integraion,
            using :func:`scipy.integrate.romberg`, is used instead.

        Returns
        -------
        complex
            The integral of the function f along the path.
        """

        args = (f, abs_tol, rel_tol, div_max, int_method)
        if args in self._integral_cache:
            return self._integral_cache[args]

        if (
            self._reverse_path is not None
            and args in self._reverse_path._integral_cache
        ):
            # if we have already computed the reverse of this path
            return -self._reverse_path._integral_cache[args]

        @functools.cache
        def integrand(t: float) -> complex:
            return f(self(t)) * self.dzdt(t)

        if int_method == "romb":
            integral = scipy.integrate.romberg(
                integrand,
                0,
                1,
                tol=abs_tol,
                rtol=rel_tol,
                divmax=div_max,
            )
        elif int_method == "quad":
            integral = scipy.integrate.quad(
                integrand, 0, 1, epsabs=abs_tol, epsrel=rel_tol, complex_func=True
            )[0]
        else:
            raise ValueError("int_method must be either 'romb' or 'quad'")

        if np.isnan(integral):
            raise RuntimeError(
                f"The integral along the segment {self} is NaN. This is most "
                "likely due to a root being on or very close to the path of "
                "integration."
            )

        self._integral_cache[args] = integral
        return integral

    def show(self, save_file: str | None = None, **plot_kwargs) -> None:
        """
        Shows the path as a 2D plot in the complex plane.  Requires
        Matplotlib.

        Parameters
        ----------
        save_file : str (optional)
            If given then the plot will be saved to disk with name
            'save_file'.  If save_file=None the plot is shown on-screen.
        **plot_kwargs
            Other key word args are passed to :meth:`~cxroots.Paths.ComplexPath.plot`
        """
        import matplotlib.pyplot as plt

        self.plot(**plot_kwargs)

        if save_file is not None:
            plt.savefig(save_file, bbox_inches="tight")
            plt.close()
        else:
            plt.show()


class ComplexLine(ComplexPath):
    r"""
    A straight line :math:`z` in the complex plane from points a to b
    parameterised by

    ..math::

        z(t) = a + (b-a)t, \quad 0\leq t \leq 1
    """

    def __init__(self, a: complex, b: complex):
        self.a, self.b = a, b
        super().__init__()

    def __str__(self):
        return (
            f"ComplexLine from {self.a.real:.3f}+{self.a.imag:.3f}i to "
            f"{self.b.real:.3f}+{self.b.imag:.3f}i"
        )

    @overload
    def __call__(self, t: float) -> complex: ...

    @overload
    def __call__(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def __call__(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        r"""
        The function :math:`z(t) = a + (b-a)t`.

        Parameters
        ----------
        t : float
            A real number :math:`0\leq t \leq 1`.

        Returns
        -------
        complex
            A point on the line in the complex plane.
        """
        return self.a + t * (self.b - self.a)

    @overload
    def dzdt(self, t: float) -> complex: ...

    @overload
    def dzdt(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def dzdt(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        """
        The derivative of the parameterised curve in the complex plane, z, with
        respect to the parameterization parameter, t.
        """
        return self.b - self.a

    def distance(self, z: complex) -> float:
        """
        Distance from the point z to the closest point on the line.

        Parameters
        ----------
        z : complex

        Returns
        -------
        float
            The distance from z to the point on the line which is
            closest to z.
        """
        # the projection of the point z onto the line a -> b is where
        # the parameter t is
        d = self.b - self.a
        t = ((z - self.a) * d.conjugate()).real / (d * d.conjugate())

        # but the line segment only has 0 <= t <= 1
        t = np.clip(t.real, 0, 1)

        # so the point on the line segment closest to z is
        c = self(t)
        return abs(c - z)


class ComplexArc(ComplexPath):
    r"""
    A circular arc :math:`z` with center z0, radius R, initial angle t0
    and change of angle dt.  The arc is parameterised by

    ..math::

        z(t) = R e^{i(t0 + t dt)} + z0, \quad 0\leq t \leq 1

    Parameters
    ----------
    z0 : complex
    R : float
    t0 : float
    dt : float
    """

    def __init__(self, z0: complex, R: float, t0: float, dt: float):  # noqa: N803
        self.z0, self.R, self.t0, self.dt = z0, R, t0, dt
        super().__init__()

    def __str__(self):
        return (
            f"ComplexArc: z0={self.z0:.3f}, R={self.R:.3f}, t0={self.t0:.3f}, "
            f"dt={self.dt:.3f}"
        )

    @overload
    def __call__(self, t: float) -> complex: ...

    @overload
    def __call__(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def __call__(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        r"""
        The function :math:`z(t) = R e^{i(t_0 + t dt)} + z_0`.

        Parameters
        ----------
        t : float
            A real number :math:`0\leq t \leq 1`.

        Returns
        -------
        complex
            A point on the arc in the complex plane.
        """
        return self.R * np.exp(1j * (self.t0 + t * self.dt)) + self.z0

    @overload
    def dzdt(self, t: float) -> complex: ...

    @overload
    def dzdt(self, t: npt.NDArray[np.float_]) -> npt.NDArray[np.complex_]: ...

    def dzdt(
        self, t: float | npt.NDArray[np.float_]
    ) -> complex | npt.NDArray[np.complex_]:
        """
        The derivative of the parameterised curve in the complex plane, z, with
        respect to the parameterization parameter, t.
        """
        return 1j * self.dt * self.R * np.exp(1j * (self.t0 + t * self.dt))

    def distance(self, z: complex) -> float:
        """
        Distance from the point z to the closest point on the arc.

        Parameters
        ----------
        z : complex

        Returns
        -------
        float
            The distance from z to the point on the arc which is closest
            to z.
        """
        theta = np.angle(z - self.z0)  # np.angle maps to (-pi,pi]
        theta = (theta - self.t0) % (2 * pi) + self.t0  # put theta in [t0,t0+2pi)

        if (self.dt > 0 and self.t0 < theta < self.t0 + self.dt) or (
            self.dt < 0 and self.t0 + self.dt < theta - 2 * pi < self.t0
        ):
            # the closest point to z lies on the arc
            return abs(self.R * np.exp(1j * theta) + self.z0 - z)
        else:
            # the closest point to z is one of the endpoints
            return min(abs(self(0) - z), abs(self(1) - z))



import numpy as np
import pytest

from cxroots import Circle, Rectangle
from cxroots.tests.approx_equal import roots_approx_equal

funcs = [
    pytest.param(
        [3, 3.01, 3.02, 8, 8.02, 8 + 0.01j], [1, 1, 1, 1, 1, 1], id="cluster_10^-2"
    ),
    pytest.param(
        [3, 3.001, 3.002, 8, 8.002, 8 + 0.001j], [1, 1, 1, 1, 1, 1], id="cluster_10^-3"
    ),
    pytest.param(
        [3, 3.0001, 3.0002, 8, 8.0002, 8 + 0.0001j],
        [1, 1, 1, 1, 1, 1],
        id="cluster_10^-4",
        marks=pytest.mark.slow,
    ),
    pytest.param(
        [3, 3.00001, 3.00002, 8, 8.00002, 8 + 0.00001j],
        [1, 1, 1, 1, 1, 1],
        id="cluster_10^-5",
        marks=[
            pytest.mark.slow,
            pytest.mark.xfail(reason="Cluster of roots too tight"),
        ],
    ),
]

contours = [
    pytest.param(Rectangle([2, 9], [-1, 1]), id="rect"),
    pytest.param(Circle(0, 8.5), id="circle"),
]


@pytest.mark.parametrize("contour", contours)
@pytest.mark.parametrize("roots,multiplicities", funcs)
def test_rootfinding_df(contour, roots, multiplicities):
    def f(z):
        return np.prod([z - r for r in roots], axis=0)

    def df(z):
        return np.sum(
            [
                np.prod([z - r for r in np.delete(roots, i)], axis=0)
                for i in range(len(roots))
            ],
            axis=0,
        )

    roots_approx_equal(contour.roots(f, df), (roots, multiplicities))


@pytest.mark.slow
@pytest.mark.parametrize("contour", contours)
@pytest.mark.parametrize("roots,multiplicities", funcs)
def test_rootfinding_f(contour, roots, multiplicities):
    def f(z):
        return np.prod([z - r for r in roots], axis=0)

    roots_approx_equal(contour.roots(f), (roots, multiplicities))



import numpy as np
import pytest
from numpy import cos, sin

from cxroots.derivative import central_diff


def test_central_diff():
    def f(z):
        return z**10 - 2 * z**5 + sin(z) * cos(z / 2)

    def df(z):
        return 10 * (z**9 - z**4) + cos(z) * cos(z / 2) - 0.5 * sin(z) * sin(z / 2)

    z = np.array([-1.234, 0.3 + 1j, 0.1j, -0.9 - 0.5j])

    approx_df = central_diff(f)

    assert approx_df(z) == pytest.approx(df(z), abs=1e-8)
